# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testtools

from tempest.lib import decorators

from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base

CONF = config.CONF


class NetworkBasicTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    # Default to ipv4.
    _ip_version = 4

    @decorators.idempotent_id('de07fe0a-e955-449e-b48b-8641c14cd52e')
    def test_basic_instance(self):
        self.setup_network_and_server()
        self.check_connectivity(self.fip['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

    @testtools.skipUnless(
        CONF.neutron_plugin_options.global_ip_address,
        'Global IP address is not defined.')
    @decorators.idempotent_id('49609189-3a0e-43c7-832e-a7e114aad1c9')
    def test_ping_global_ip_from_vm_with_fip(self):
        self.setup_network_and_server()
        server_ssh_client = ssh.Client(self.fip['floating_ip_address'],
                                       CONF.validation.image_ssh_user,
                                       pkey=self.keypair['private_key'])
        self.check_remote_connectivity(
            server_ssh_client,
            CONF.neutron_plugin_options.global_ip_address,
            ping_count=1)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron_lib import constants
from tempest.common import utils
from tempest.lib import decorators
from tempest.lib import exceptions

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config


class TestRevisions(base.BaseAdminNetworkTest):

    required_extensions = ['standard-attr-revisions']

    @decorators.idempotent_id('4a26a4be-9c53-483c-bc50-b53f1db10ac6')
    def test_update_network_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        self.assertIn('revision_number', net)
        updated = self.client.update_network(net['id'], name='newnet')
        self.assertGreater(updated['network']['revision_number'],
                           net['revision_number'])

    @decorators.idempotent_id('4a26a4be-9c53-483c-bc50-b11111113333')
    def test_update_network_constrained_by_revision(self):
        net = self.create_network()
        current = net['revision_number']
        stale = current - 1
        # using a stale number should fail
        self.assertRaises(
            exceptions.PreconditionFailed,
            self.client.update_network,
            net['id'], name='newnet',
            headers={'If-Match': 'revision_number=%s' % stale}
        )

        # using current should pass. in case something is updating the network
        # on the server at the same time, we have to re-read and update to be
        # safe
        for i in range(100):
            current = (self.client.show_network(net['id'])
                       ['network']['revision_number'])
            try:
                self.client.update_network(
                    net['id'], name='newnet',
                    headers={'If-Match': 'revision_number=%s' % current})
            except exceptions.UnexpectedResponseCode:
                continue
            break
        else:
            self.fail("Failed to update network after 100 tries.")

    @decorators.idempotent_id('cac7ecde-12d5-4331-9a03-420899dea077')
    def test_update_port_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        self.assertIn('revision_number', port)
        updated = self.client.update_port(port['id'], name='newport')
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])

    @decorators.idempotent_id('c1c4fa41-8e89-44d0-9bfc-409f3b66dc57')
    def test_update_subnet_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        self.addCleanup(self.client.delete_subnet, subnet['id'])
        self.assertIn('revision_number', subnet)
        updated = self.client.update_subnet(subnet['id'], name='newsub')
        self.assertGreater(updated['subnet']['revision_number'],
                           subnet['revision_number'])

    @decorators.idempotent_id('e8c5d7db-2b8d-4615-a476-6e537437c4f2')
    def test_update_subnetpool_bumps_revision(self):
        sp = self.create_subnetpool('subnetpool', default_prefixlen=24,
                                    prefixes=['10.0.0.0/8'])
        self.addCleanup(self.client.delete_subnetpool, sp['id'])
        self.assertIn('revision_number', sp)
        updated = self.admin_client.update_subnetpool(sp['id'], name='sp2')
        self.assertGreater(updated['subnetpool']['revision_number'],
                           sp['revision_number'])

    @decorators.idempotent_id('e8c5d7db-2b8d-4567-a326-6e123437c4d1')
    def test_update_subnet_bumps_network_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        updated = self.client.show_network(net['id'])
        self.assertGreater(updated['network']['revision_number'],
                           net['revision_number'])
        self.client.delete_subnet(subnet['id'])
        updated2 = self.client.show_network(net['id'])
        self.assertGreater(updated2['network']['revision_number'],
                           updated['network']['revision_number'])

    @decorators.idempotent_id('6c256f71-c929-4200-b3dc-4e1843506be5')
    @utils.requires_ext(extension="security-group", service="network")
    def test_update_sg_group_bumps_revision(self):
        security_group = self.create_security_group()
        self.assertIn('revision_number', security_group)
        updated_security_group = self.client.update_security_group(
            security_group['id'], name='new_sg_name')['security_group']
        self.assertGreater(updated_security_group['revision_number'],
                           security_group['revision_number'])

    @decorators.idempotent_id('6489632f-8550-4453-a674-c98849742967')
    @utils.requires_ext(extension="security-group", service="network")
    def test_update_port_sg_binding_bumps_revision(self):
        network = self.create_network()
        port = self.create_port(network)

        security_group = self.create_security_group()
        updated_port = self.client.update_port(
            port['id'], security_groups=[security_group['id']])['port']
        self.assertGreater(updated_port['revision_number'],
                           port['revision_number'])

        updated_port2 = self.client.update_port(
            port['id'], security_groups=[])['port']
        self.assertGreater(updated_port2['revision_number'],
                           updated_port['revision_number'])

    @decorators.idempotent_id('29c7ab2b-d1d8-425d-8cec-fcf632960f22')
    @utils.requires_ext(extension="security-group", service="network")
    def test_update_sg_rule_bumps_sg_revision(self):
        security_group = self.create_security_group()

        security_group_rule = self.create_security_group_rule(
            security_group=security_group,
            protocol=constants.PROTO_NAME_TCP,
            direction=constants.INGRESS_DIRECTION,
            port_range_min=60,
            port_range_max=70)
        updated_security_group = self.client.show_security_group(
            security_group['id'])['security_group']
        self.assertGreater(updated_security_group['revision_number'],
                           security_group['revision_number'])

        self.client.delete_security_group_rule(security_group_rule['id'])
        updated_security_group2 = self.client.show_security_group(
            security_group['id'])['security_group']
        self.assertGreater(updated_security_group2['revision_number'],
                           updated_security_group['revision_number'])

    @decorators.idempotent_id('db70c285-0365-4fac-9f55-2a0ad8cf55a8')
    @utils.requires_ext(extension="allowed-address-pairs", service="network")
    def test_update_allowed_address_pairs_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        updated = self.client.update_port(
            port['id'], allowed_address_pairs=[{'ip_address': '1.1.1.1/32'}])
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])
        updated2 = self.client.update_port(
            port['id'], allowed_address_pairs=[])
        self.assertGreater(updated2['port']['revision_number'],
                           updated['port']['revision_number'])

    @decorators.idempotent_id('a21ec3b4-3569-4b77-bf29-4177edaa2df5')
    @utils.requires_ext(extension="extra_dhcp_opt", service="network")
    def test_update_extra_dhcp_opt_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        opts = [{'opt_value': 'pxelinux.0', 'opt_name': 'bootfile-name'}]
        updated = self.client.update_port(port['id'], extra_dhcp_opts=opts)
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])
        opts[0]['opt_value'] = 'pxelinux.77'
        updated2 = self.client.update_port(
            port['id'], extra_dhcp_opts=opts)
        self.assertGreater(updated2['port']['revision_number'],
                           updated['port']['revision_number'])

    @decorators.idempotent_id('40ba648f-f374-4c29-a5b7-489dd5a38a4e')
    @utils.requires_ext(extension="dns-integration", service="network")
    def test_update_dns_domain_bumps_revision(self):
        net = self.create_network(dns_domain='example.test.')
        self.addCleanup(self.client.delete_network, net['id'])
        updated = self.client.update_network(net['id'], dns_domain='exa.test.')
        self.assertGreater(updated['network']['revision_number'],
                           net['revision_number'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        updated = self.client.update_port(port['id'], dns_name='port1')
        if not updated['port']['dns_name']:
            self.skipTest("Server does not have DNS domain configured.")
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])
        updated2 = self.client.update_port(port['id'], dns_name='')
        self.assertGreater(updated2['port']['revision_number'],
                           updated['port']['revision_number'])

    @decorators.idempotent_id('8482324f-cf59-4d73-b98e-d37119255300')
    @utils.requires_ext(extension="router", service="network")
    @utils.requires_ext(extension="extraroute", service="network")
    def test_update_router_extra_routes_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        self.addCleanup(self.client.delete_subnet, subnet['id'])
        subgateway = netaddr.IPAddress(subnet['gateway_ip'])
        router = self.create_router(router_name='test')
        self.addCleanup(self.client.delete_router, router['id'])
        self.create_router_interface(router['id'], subnet['id'])
        self.addCleanup(
            self.client.remove_router_interface_with_subnet_id,
            router['id'],
            subnet['id'])
        router = self.client.show_router(router['id'])['router']
        updated = self.client.update_extra_routes(
            router['id'], str(subgateway + 1), '2.0.0.0/24')
        self.assertGreater(updated['router']['revision_number'],
                           router['revision_number'])
        updated2 = self.client.delete_extra_routes(router['id'])
        self.assertGreater(updated2['router']['revision_number'],
                           updated['router']['revision_number'])

    @decorators.idempotent_id('6bd18702-e25a-4b4b-8c0c-680113533511')
    @utils.requires_ext(extension="subnet-service-types", service="network")
    def test_update_subnet_service_types_bumps_revisions(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        self.addCleanup(self.client.delete_subnet, subnet['id'])
        updated = self.client.update_subnet(
            subnet['id'], service_types=['compute:'])
        self.assertGreater(updated['subnet']['revision_number'],
                           subnet['revision_number'])
        updated2 = self.client.update_subnet(
            subnet['id'], service_types=[])
        self.assertGreater(updated2['subnet']['revision_number'],
                           updated['subnet']['revision_number'])

    @decorators.idempotent_id('9c83105c-9973-45ff-9ca2-e66d64700abe')
    @utils.requires_ext(extension="port-security", service="network")
    def test_update_port_security_bumps_revisions(self):
        net = self.create_network(port_security_enabled=False)
        self.addCleanup(self.client.delete_network, net['id'])
        updated = self.client.update_network(net['id'],
                                             port_security_enabled=True)
        self.assertGreater(updated['network']['revision_number'],
                           net['revision_number'])
        updated2 = self.client.update_network(net['id'],
                                              port_security_enabled=False)
        self.assertGreater(updated2['network']['revision_number'],
                           updated['network']['revision_number'])
        port = self.create_port(net, port_security_enabled=False)
        self.addCleanup(self.client.delete_port, port['id'])
        updated = self.client.update_port(port['id'],
                                          port_security_enabled=True)
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])
        updated2 = self.client.update_port(port['id'],
                                           port_security_enabled=False)
        self.assertGreater(updated2['port']['revision_number'],
                           updated['port']['revision_number'])

    @decorators.idempotent_id('68d5ac3a-11a1-4847-8e2e-5843c043d89b')
    @utils.requires_ext(extension="binding", service="network")
    def test_portbinding_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        port = self.admin_client.update_port(
            port['id'], **{'binding:host_id': 'badhost1'})['port']
        updated = self.admin_client.update_port(
            port['id'], **{'binding:host_id': 'badhost2'})['port']
        self.assertGreater(updated['revision_number'],
                           port['revision_number'])

    @decorators.idempotent_id('4a37bde9-1975-47e0-9b8c-2c9ca36415b0')
    @utils.requires_ext(extension="router", service="network")
    def test_update_router_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        self.addCleanup(self.client.delete_subnet, subnet['id'])
        router = self.create_router(router_name='test')
        self.addCleanup(self.client.delete_router, router['id'])
        self.assertIn('revision_number', router)
        rev1 = router['revision_number']
        router = self.client.update_router(router['id'],
                                           name='test2')['router']
        self.assertGreater(router['revision_number'], rev1)
        self.create_router_interface(router['id'], subnet['id'])
        self.addCleanup(
            self.client.remove_router_interface_with_subnet_id,
            router['id'],
            subnet['id'])
        updated = self.client.show_router(router['id'])['router']
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])

    @decorators.idempotent_id('9de71ebc-f5df-4cd0-80bc-60299fce3ce9')
    @utils.requires_ext(extension="router", service="network")
    @utils.requires_ext(extension="standard-attr-description",
                        service="network")
    def test_update_floatingip_bumps_revision(self):
        ext_id = config.CONF.network.public_network_id
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        subnet = self.create_subnet(net)
        self.addCleanup(self.client.delete_subnet, subnet['id'])
        router = self.create_router('test', external_network_id=ext_id)
        self.addCleanup(self.client.delete_router, router['id'])
        self.create_router_interface(router['id'], subnet['id'])
        self.addCleanup(
            self.client.remove_router_interface_with_subnet_id,
            router['id'],
            subnet['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        body = self.create_floatingip(port=port, description='d1')
        self.assertIn('revision_number', body)
        b2 = self.client.update_floatingip(body['id'], description='d2')
        self.assertGreater(b2['floatingip']['revision_number'],
                           body['revision_number'])
        # disassociate
        self.client.update_floatingip(b2['floatingip']['id'], port_id=None)

    @decorators.idempotent_id('afb6486c-41b5-483e-a500-3c506f4deb49')
    @utils.requires_ext(extension="router", service="network")
    @utils.requires_ext(extension="l3-ha", service="network")
    def test_update_router_extra_attributes_bumps_revision(self):
        # OVN driver doesn't accept ha=False (all OVN routers support HA)
        if self.is_driver_ovn:
            raise self.skipException("Test not meant for OVN driver")
        # updates from CVR to CVR-HA are supported on every release,
        # but only the admin can forcibly create a non-HA router
        router_args = {'tenant_id': self.client.project_id,
                       'ha': False}
        router = self.admin_client.create_router('r1', True,
            **router_args)['router']
        self.addCleanup(self.client.delete_router, router['id'])
        self.assertIn('revision_number', router)
        rev1 = router['revision_number']
        router = self.admin_client.update_router(
            router['id'], admin_state_up=False)['router']
        self.assertGreater(router['revision_number'], rev1)
        self.admin_client.update_router(router['id'], ha=True)['router']
        updated = self.client.show_router(router['id'])['router']
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])

    @decorators.idempotent_id('90743b00-b0e2-40e4-9524-1c884fe3ef23')
    @utils.requires_ext(extension="external-net", service="network")
    @utils.requires_ext(extension="auto-allocated-topology", service="network")
    @utils.requires_ext(extension="subnet_allocation", service="network")
    @utils.requires_ext(extension="router", service="network")
    def test_update_external_network_bumps_revision(self):
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        self.assertIn('revision_number', net)
        updated = self.admin_client.update_network(net['id'],
                                                   **{'router:external': True})
        self.assertGreater(updated['network']['revision_number'],
                           net['revision_number'])

    @decorators.idempotent_id('5af6450a-0f61-49c3-b628-38db77c7b856')
    @utils.requires_ext(extension="qos", service="network")
    def test_update_qos_port_policy_binding_bumps_revision(self):
        policy = self.create_qos_policy(name='port-policy', shared=False)
        net = self.create_network()
        self.addCleanup(self.client.delete_network, net['id'])
        port = self.create_port(net)
        self.addCleanup(self.client.delete_port, port['id'])
        updated = self.admin_client.update_port(
            port['id'], qos_policy_id=policy['id'])
        self.assertGreater(updated['port']['revision_number'],
                           port['revision_number'])

    @decorators.idempotent_id('817da343-c6e4-445c-9519-a621f124dfbe')
    @utils.requires_ext(extension="qos", service="network")
    def test_update_qos_network_policy_binding_bumps_revision(self):
        policy = self.create_qos_policy(name='network-policy', shared=False)
        network = self.create_network()
        self.addCleanup(self.client.delete_network, network['id'])
        updated = self.admin_client.update_network(
            network['id'], qos_policy_id=policy['id'])
        self.assertGreater(updated['network']['revision_number'],
                           network['revision_number'])



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random

from neutron_lib import constants
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import decorators
from tempest.lib import exceptions
import testtools

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin.api import base_security_groups
from oslo_log import log

LOG = log.getLogger(__name__)


class BaseSecGroupTest(base.BaseAdminNetworkTest):

    required_extensions = ['security-group']

    def _test_create_list_update_show_delete_security_group(self):
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        security_group = self.create_security_group(**sg_kwargs)

        # List security groups and verify if created group is there in response
        security_groups = self.client.list_security_groups()['security_groups']
        self.assertIn(security_group['id'],
                      {sg['id'] for sg in security_groups})

        # Update the security group
        new_name = data_utils.rand_name('security')
        new_description = data_utils.rand_name('security-description')
        updated_security_group = self.client.update_security_group(
            security_group['id'], name=new_name,
            description=new_description)['security_group']

        # Verify if security group is updated
        self.assertEqual(updated_security_group['name'], new_name)
        self.assertEqual(updated_security_group['description'],
                         new_description)

        # Show details of the updated security group
        observed_security_group = self.client.show_security_group(
            security_group['id'])['security_group']
        self.assertEqual(observed_security_group['name'], new_name)
        self.assertEqual(observed_security_group['description'],
                         new_description)

    def _test_show_security_group_contains_all_rules(self):
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        security_group = self.create_security_group(**sg_kwargs)
        protocol = random.choice(list(base_security_groups.V4_PROTOCOL_NAMES))
        security_group_rule = self.create_security_group_rule(
            security_group=security_group,
            project={'id': self.admin_client.project_id},
            client=self.admin_client,
            protocol=protocol,
            direction=constants.INGRESS_DIRECTION)

        observed_security_group = self.client.show_security_group(
            security_group['id'])['security_group']
        observerd_security_group_rules_ids = [
            sgr['id'] for sgr in
            observed_security_group['security_group_rules']]
        self.assertIn(
            security_group_rule['id'], observerd_security_group_rules_ids)

    def _test_list_security_group_rules_contains_all_rules(self):
        """Test list security group rules.

        This test checks if all SG rules which belongs to the tenant OR
        which belongs to the tenant's security group are listed.
        """
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        security_group = self.create_security_group(**sg_kwargs)
        protocol = random.choice(list(base_security_groups.V4_PROTOCOL_NAMES))
        security_group_rule = self.create_security_group_rule(
            security_group=security_group,
            project={'id': self.admin_client.project_id},
            client=self.admin_client,
            protocol=protocol,
            direction=constants.INGRESS_DIRECTION)

        # Create also other SG with some custom rule to check that regular user
        # can't see this rule
        sg_kwargs = {
            'project': {'id': self.admin_client.project_id},
            'client': self.admin_client
        }
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        admin_security_group = self.create_security_group(**sg_kwargs)
        admin_security_group_rule = self.create_security_group_rule(
            security_group=admin_security_group,
            project={'id': self.admin_client.project_id},
            client=self.admin_client,
            protocol=protocol,
            direction=constants.INGRESS_DIRECTION)

        rules = self.client.list_security_group_rules()['security_group_rules']
        rules_ids = [rule['id'] for rule in rules]
        self.assertIn(security_group_rule['id'], rules_ids)
        self.assertNotIn(admin_security_group_rule['id'], rules_ids)

    def _test_create_bulk_sec_groups(self):
        # Creates 2 sec-groups in one request
        sec_nm = [data_utils.rand_name('secgroup'),
                  data_utils.rand_name('secgroup')]
        body = self.client.create_bulk_security_groups(
            sec_nm, stateless=self.stateless_sg)
        created_sec_grps = body['security_groups']
        self.assertEqual(2, len(created_sec_grps))
        for secgrp in created_sec_grps:
            self.addCleanup(self.client.delete_security_group,
                            secgrp['id'])
            self.assertIn(secgrp['name'], sec_nm)
            self.assertIsNotNone(secgrp['id'])

    def _test_create_sec_groups_with_the_same_name(self):
        same_name_sg_number = 5
        sg_name = 'sg_zahlabut'
        sg_names = [sg_name] * same_name_sg_number
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        for name in sg_names:
            sg_kwargs['name'] = name
            self.create_security_group(**sg_kwargs)
        sec_groups = [item['id'] for item in
                      self.client.list_security_groups(
                          name=sg_name)['security_groups']]
        self.assertEqual(
            same_name_sg_number, len(set(sec_groups)),
            'Failed - expected number of groups with the same name'
            ' is: {}'.format(same_name_sg_number))


class StatefulSecGroupTest(BaseSecGroupTest):

    stateless_sg = False

    @decorators.idempotent_id('bfd128e5-3c92-44b6-9d66-7fe29d22c802')
    def test_create_list_update_show_delete_security_group(self):
        self._test_create_list_update_show_delete_security_group()

    @decorators.idempotent_id('1fff0d57-bb6c-4528-9c1d-2326dce1c087')
    def test_show_security_group_contains_all_rules(self):
        self._test_show_security_group_contains_all_rules()

    @decorators.idempotent_id('b5923b1a-4d33-44e1-af25-088dcb55b02b')
    def test_list_security_group_rules_contains_all_rules(self):
        self._test_list_security_group_rules_contains_all_rules()

    @decorators.idempotent_id('7c0ecb10-b2db-11e6-9b14-000c29248b0d')
    def test_create_bulk_sec_groups(self):
        self._test_create_bulk_sec_groups()

    @decorators.idempotent_id('e93f33d8-57ea-11eb-b69b-74e5f9e2a801')
    def test_create_sec_groups_with_the_same_name(self):
        self._test_create_sec_groups_with_the_same_name()


class StatelessSecGroupTest(BaseSecGroupTest):

    required_extensions = ['security-group', 'stateful-security-group']
    stateless_sg = True

    @decorators.idempotent_id('0214d58a-2177-47e1-af83-dcd45c024829')
    def test_create_list_update_show_delete_security_group(self):
        self._test_create_list_update_show_delete_security_group()

    @decorators.idempotent_id('ddbc0e4c-840f-44ab-8718-0b95b7c7b575')
    def test_show_security_group_contains_all_rules(self):
        self._test_show_security_group_contains_all_rules()

    @decorators.idempotent_id('cdf3a63a-08fe-4091-bab4-62180847990f')
    def test_list_security_group_rules_contains_all_rules(self):
        self._test_list_security_group_rules_contains_all_rules()

    @decorators.idempotent_id('b33e612e-65f0-467b-9bf2-b5b2ce67f72f')
    def test_create_bulk_sec_groups(self):
        self._test_create_bulk_sec_groups()

    @decorators.idempotent_id('a6896935-db18-413d-95f5-4f465e0e2209')
    def test_create_sec_groups_with_the_same_name(self):
        self._test_create_sec_groups_with_the_same_name()

    @decorators.idempotent_id('0a6c1476-3d1a-11ec-b0ec-0800277ac3d9')
    def test_stateless_security_group_update(self):
        security_group = self.create_security_group(stateful=True)

        # List security groups and verify if created group is there in response
        security_groups = self.client.list_security_groups()['security_groups']
        found = False
        for sg in security_groups:
            if sg['id'] == security_group['id']:
                found = True
                break
        self.assertTrue(found)
        self.assertTrue(sg['stateful'])

        # Switch to stateless
        updated_security_group = self.client.update_security_group(
            security_group['id'], stateful=False)['security_group']

        # Verify if security group is updated
        self.assertFalse(updated_security_group['stateful'])

        observed_security_group = self.client.show_security_group(
            security_group['id'])['security_group']
        self.assertFalse(observed_security_group['stateful'])

        # Switch back to stateful
        updated_security_group = self.client.update_security_group(
            security_group['id'], stateful=True)['security_group']

        # Verify if security group is stateful again
        self.assertTrue(updated_security_group['stateful'])

        observed_security_group = self.client.show_security_group(
            security_group['id'])['security_group']
        self.assertTrue(observed_security_group['stateful'])


class BaseSecGroupQuota(base.BaseAdminNetworkTest):

    def setUp(self):
        super().setUp()
        # NOTE(slaweq): we don't know exactly how many rule templates may be
        # created in the neutron db and used for every SG so, as in this test
        # class we are checking quotas of SG, not SG rules, lets set quota for
        # SG rules to -1
        self._set_sg_rules_quota(-1)

    def _create_max_allowed_sg_amount(self):
        sg_amount = self._get_sg_amount()
        sg_quota = self._get_sg_quota()
        sg_to_create = sg_quota - sg_amount
        self._create_security_groups(sg_to_create)

    def _create_security_groups(self, amount):
        for _ in range(amount):
            sg = self.create_security_group()
            self.addCleanup(self.delete_security_group, sg)

    def _increase_sg_quota(self):
        sg_quota = self._get_sg_quota()
        new_sg_quota = 2 * sg_quota
        self._set_sg_quota(new_sg_quota)
        self.assertEqual(self._get_sg_quota(), new_sg_quota,
                         "Security group quota wasn't changed correctly")

    def _decrease_sg_quota(self):
        sg_quota = self._get_sg_quota()
        new_sg_quota = sg_quota // 2
        self._set_sg_quota(new_sg_quota)
        self.assertEqual(self._get_sg_quota(), new_sg_quota,
                         "Security group quota wasn't changed correctly")

    def _set_quota(self, val, resource):
        res_quota = self._get_quota(resource)
        project_id = self.client.project_id
        self.admin_client.update_quotas(project_id, **{resource: val})
        self.addCleanup(self.admin_client.update_quotas,
                        project_id, **{resource: res_quota})

    def _get_quota(self, resource):
        project_id = self.client.project_id
        quotas = self.admin_client.show_quotas(project_id)
        return quotas['quota'][resource]

    def _set_sg_quota(self, val):
        return self._set_quota(val, 'security_group')

    def _get_sg_quota(self):
        return self._get_quota('security_group')

    def _get_sg_amount(self):
        project_id = self.client.project_id
        filter_query = {'project_id': project_id}
        security_groups = self.client.list_security_groups(**filter_query)
        return len(security_groups['security_groups'])

    def _set_sg_rules_quota(self, val):
        return self._set_quota(val, 'security_group_rule')


class SecGroupQuotaTest(BaseSecGroupQuota):

    credentials = ['primary', 'admin']
    required_extensions = ['security-group', 'quotas']

    @decorators.idempotent_id('1826aa02-090d-4717-b43a-50ee449b02e7')
    def test_sg_quota_values(self):
        values = [-1, 0, 10, 2147483647]
        for value in values:
            self._set_sg_quota(value)
            self.assertEqual(value, self._get_sg_quota())

    @decorators.idempotent_id('df7981fb-b83a-4779-b13e-65494ef44a72')
    def test_max_allowed_sg_amount(self):
        self._create_max_allowed_sg_amount()
        self.assertEqual(self._get_sg_quota(), self._get_sg_amount())

    @decorators.idempotent_id('623d909c-6ef8-43d6-93ee-97086e2651e8')
    def test_sg_quota_increased(self):
        self._create_max_allowed_sg_amount()
        self._increase_sg_quota()
        self._create_max_allowed_sg_amount()
        self.assertEqual(self._get_sg_quota(), self._get_sg_amount(),
                         "Amount of security groups doesn't match quota")

    @decorators.idempotent_id('ba95676c-8d9a-4482-b4ec-74d51a4602a6')
    def test_sg_quota_decrease_less_than_created(self):
        self._create_max_allowed_sg_amount()
        self._decrease_sg_quota()

    @decorators.idempotent_id('d43cf1a7-aa7e-4c41-9340-627a1a6ab961')
    def test_create_sg_when_quota_disabled(self):
        sg_amount = self._get_sg_amount()
        self._set_sg_quota(-1)
        self._create_security_groups(10)
        new_sg_amount = self._get_sg_amount()
        self.assertGreater(new_sg_amount, sg_amount)


class BaseSecGroupRulesQuota(base.BaseAdminNetworkTest):

    def _create_max_allowed_sg_rules_amount(self, port_index=1):
        sg_rules_amount = self._get_sg_rules_amount()
        sg_rules_quota = self._get_sg_rules_quota()
        sg_rules_to_create = sg_rules_quota - sg_rules_amount
        port_index += sg_rules_to_create
        self._create_security_group_rules(sg_rules_to_create,
                                          port_index=port_index)

    def _create_security_group_rules(self, amount, port_index=1):
        for i in range(amount):
            ingress_rule = self.create_security_group_rule(**{
                'project_id': self.client.project_id,
                'direction': 'ingress',
                'port_range_max': port_index + i,
                'port_range_min': port_index + i,
                'protocol': 'tcp'})
            self.addCleanup(
                self.client.delete_security_group_rule, ingress_rule['id'])

    def _increase_sg_rules_quota(self):
        sg_rules_quota = self._get_sg_rules_quota()
        new_sg_rules_quota = 2 * sg_rules_quota
        self._set_sg_rules_quota(new_sg_rules_quota)
        self.assertGreater(self._get_sg_rules_quota(), sg_rules_quota,
                           "Security group rule quota wasnt changed correctly")
        return new_sg_rules_quota

    def _decrease_sg_rules_quota(self):
        sg_rules_quota = self._get_sg_rules_quota()
        new_sg_rules_quota = sg_rules_quota // 2
        self._set_sg_rules_quota(new_sg_rules_quota)
        return new_sg_rules_quota

    def _set_sg_rules_quota(self, val):
        project_id = self.client.project_id
        self.admin_client.update_quotas(project_id,
                                        **{'security_group_rule': val})
        LOG.info('Trying to update security group rule quota {} '.format(val))

    def _get_sg_rules_quota(self):
        project_id = self.client.project_id
        quotas = self.admin_client.show_quotas(project_id)
        return quotas['quota']['security_group_rule']

    def _get_sg_rules_amount(self):
        project_id = self.client.project_id
        filter_query = {'project_id': project_id}
        security_group_rules = self.client.list_security_group_rules(
                **filter_query)
        return len(security_group_rules['security_group_rules'])


class SecGroupRulesQuotaTest(BaseSecGroupRulesQuota):

    credentials = ['primary', 'admin']
    required_extensions = ['security-group', 'quotas']

    def setUp(self):
        super(SecGroupRulesQuotaTest, self).setUp()
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        self.admin_client.reset_quotas, self.client.project_id)
        self._set_sg_rules_quota(10)

    @decorators.idempotent_id('77ec038c-5638-11ea-8e2d-0242ac130003')
    def test_sg_rules_quota_increased(self):
        """Test security group rules quota increased.

        This test checks if it is possible to increase the SG rules Quota
        value and creates security group rules according to new quota value.
        """
        self._create_max_allowed_sg_rules_amount()
        new_quota = self._increase_sg_rules_quota()
        port_index = new_quota
        self._create_max_allowed_sg_rules_amount(port_index)
        quota_set = self._get_sg_rules_quota()
        self.assertEqual(quota_set, self._get_sg_rules_amount(),
                         "Amount of security groups rules doesn't match quota")

    @decorators.idempotent_id('37508c8d-270b-4b93-8007-72876a1fec38')
    def test_sg_rules_quota_values(self):
        """Test security group rules quota values.

        This test checks if it is possible to change the SG rules Quota
        values, different values.
        """
        sg_rules_quota = self._get_sg_rules_quota()
        project_id = self.client.project_id
        self.addCleanup(self.admin_client.update_quotas,
                        project_id, **{'security_group_rule': sg_rules_quota})
        values = [-1, 0, 10, 2147483647]
        for value in values:
            self._set_sg_rules_quota(value)
            self.assertEqual(value, self._get_sg_rules_quota())

    @decorators.idempotent_id('4459e066-d9c8-4a13-9e98-018f95ce2dbf')
    def test_create_sg_rules_when_quota_disabled(self):
        sg_rules_amount = self._get_sg_rules_amount()
        self._set_sg_rules_quota(-1)
        self._create_security_group_rules(10, port_index=100)
        new_sg_rules_amount = self._get_sg_rules_amount()
        self.assertGreater(new_sg_rules_amount, sg_rules_amount)

    @decorators.idempotent_id('3fc39bd6-3132-4e6f-a09c-456fb18d600c')
    def test_sg_rules_quota_decrease_less_than_created(self):
        self._create_max_allowed_sg_rules_amount()
        new_quota = self._decrease_sg_rules_quota()
        self.assertEqual(self._get_sg_rules_quota(), new_quota)


class BaseSecGroupProtocolTest(base.BaseNetworkTest):

    protocol_names = base_security_groups.V4_PROTOCOL_NAMES
    protocol_ints = base_security_groups.V4_PROTOCOL_INTS

    def _test_security_group_rule_protocols(self, protocols):
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        security_group = self.create_security_group(**sg_kwargs)
        for protocol in protocols:
            self._test_security_group_rule(
                security_group=security_group,
                protocol=str(protocol),
                direction=constants.INGRESS_DIRECTION,
                ethertype=self.ethertype)

    def _test_security_group_rule(self, security_group, **kwargs):
        security_group_rule = self.create_security_group_rule(
            security_group=security_group, **kwargs)
        observed_security_group_rule = self.client.show_security_group_rule(
            security_group_rule['id'])['security_group_rule']
        for key, value in kwargs.items():
            self.assertEqual(value, security_group_rule[key],
                             "{!r} does not match.".format(key))
            self.assertEqual(value, observed_security_group_rule[key],
                             "{!r} does not match.".format(key))


class StatefulSecGroupProtocolTest(BaseSecGroupProtocolTest):
    stateless_sg = False

    @decorators.idempotent_id('282e3681-aa6e-42a7-b05c-c341aa1e3cdf')
    def test_security_group_rule_protocol_names(self):
        self._test_security_group_rule_protocols(protocols=self.protocol_names)

    @decorators.idempotent_id('66e47f1f-20b6-4417-8839-3cc671c7afa3')
    def test_security_group_rule_protocol_ints(self):
        self._test_security_group_rule_protocols(protocols=self.protocol_ints)


class StatelessSecGroupProtocolTest(BaseSecGroupProtocolTest):
    required_extensions = ['security-group', 'stateful-security-group']
    stateless_sg = True

    @decorators.idempotent_id('3a065cdd-99bd-409f-a08e-385c6674bec2')
    def test_security_group_rule_protocol_names(self):
        self._test_security_group_rule_protocols(protocols=self.protocol_names)

    @decorators.idempotent_id('b0332b5d-6fac-49d5-a79d-ae4fe62600f7')
    def test_security_group_rule_protocol_ints(self):
        self._test_security_group_rule_protocols(protocols=self.protocol_ints)


class BaseSecGroupProtocolIPv6Test(BaseSecGroupProtocolTest):

    _ip_version = constants.IP_VERSION_6
    protocol_names = base_security_groups.V6_PROTOCOL_NAMES
    protocol_ints = base_security_groups.V6_PROTOCOL_INTS

    def _test_security_group_rule_protocol_legacy_icmpv6(self):
        # These legacy protocols can be used to create security groups,
        # but they could be shown either with their passed protocol name,
        # or a canonical-ized version, depending on the neutron version.
        # So we check against a list of possible values.
        # TODO(haleyb): Remove once these legacy names are deprecated
        protocols = {constants.PROTO_NAME_IPV6_ICMP_LEGACY:
                     constants.PROTO_NAME_IPV6_ICMP,
                     constants.PROTO_NAME_ICMP:
                     constants.PROTO_NAME_IPV6_ICMP}
        for key, value in protocols.items():
            self._test_security_group_rule_legacy(
                protocol_list=[str(key), str(value)],
                protocol=str(key),
                direction=constants.INGRESS_DIRECTION,
                ethertype=self.ethertype)

    def _test_security_group_rule_legacy(self, protocol_list, **kwargs):
        sg_kwargs = {}
        if self.stateless_sg:
            sg_kwargs['stateful'] = False
        security_group = self.create_security_group(**sg_kwargs)
        security_group_rule = self.create_security_group_rule(
            security_group=security_group, **kwargs)
        observed_security_group_rule = self.client.show_security_group_rule(
            security_group_rule['id'])['security_group_rule']
        for key, value in kwargs.items():
            if key == 'protocol':
                self.assertIn(security_group_rule[key], protocol_list,
                              "{!r} does not match.".format(key))
                self.assertIn(observed_security_group_rule[key], protocol_list,
                              "{!r} does not match.".format(key))
            else:
                self.assertEqual(value, security_group_rule[key],
                                 "{!r} does not match.".format(key))
                self.assertEqual(value, observed_security_group_rule[key],
                                 "{!r} does not match.".format(key))


class StatefulSecGroupProtocolIPv6Test(BaseSecGroupProtocolIPv6Test):
    stateless_sg = False

    @decorators.idempotent_id('c7d17b41-3b4e-4add-bb3b-6af59baaaffa')
    def test_security_group_rule_protocol_legacy_icmpv6(self):
        self._test_security_group_rule_protocol_legacy_icmpv6()


class StatelessSecGroupProtocolIPv6Test(BaseSecGroupProtocolIPv6Test):
    required_extensions = ['security-group', 'stateful-security-group']
    stateless_sg = True

    @decorators.idempotent_id('a034814e-0fa5-4437-8e6f-0d2eebd668b3')
    def test_security_group_rule_protocol_legacy_icmpv6(self):
        self._test_security_group_rule_protocol_legacy_icmpv6()


class RbacSharedSecurityGroupTest(base.BaseAdminNetworkTest):

    force_tenant_isolation = True
    credentials = ['primary', 'alt', 'admin']
    required_extensions = ['security-group', 'rbac-security-groups']

    @classmethod
    def resource_setup(cls):
        super(RbacSharedSecurityGroupTest, cls).resource_setup()
        cls.client2 = cls.os_alt.network_client

    def _create_security_group(self):
        return self.create_security_group(
            name=data_utils.rand_name('test-sg'),
            project={'id': self.admin_client.project_id})

    def _make_admin_sg_shared_to_project_id(self, project_id):
        sg = self._create_security_group()
        rbac_policy = self.admin_client.create_rbac_policy(
            object_type='security_group',
            object_id=sg['id'],
            action='access_as_shared',
            target_tenant=project_id,
        )['rbac_policy']
        return {'security_group': sg, 'rbac_policy': rbac_policy}

    @decorators.idempotent_id('2a41eb8f-2a35-11e9-bae9-acde48001122')
    def test_policy_target_update(self):
        res = self._make_admin_sg_shared_to_project_id(
            self.client.project_id)
        # change to client2
        update_res = self.admin_client.update_rbac_policy(
            res['rbac_policy']['id'], target_tenant=self.client2.project_id)
        self.assertEqual(self.client2.project_id,
                         update_res['rbac_policy']['target_tenant'])
        # make sure everything else stayed the same
        res['rbac_policy'].pop('target_tenant')
        update_res['rbac_policy'].pop('target_tenant')
        self.assertEqual(res['rbac_policy'], update_res['rbac_policy'])

    @decorators.idempotent_id('2a619a8a-2a35-11e9-90d9-acde48001122')
    def test_port_presence_prevents_policy_rbac_policy_deletion(self):
        res = self._make_admin_sg_shared_to_project_id(
            self.client2.project_id)
        sg_id = res['security_group']['id']
        net = self.create_network(client=self.client2)
        port = self.client2.create_port(
            network_id=net['id'],
            security_groups=[sg_id])['port']

        # a port with shared sg should prevent the deletion of an
        # rbac-policy required for it to be shared
        with testtools.ExpectedException(exceptions.Conflict):
            self.admin_client.delete_rbac_policy(res['rbac_policy']['id'])

        # cleanup
        self.client2.delete_port(port['id'])
        self.admin_client.delete_rbac_policy(res['rbac_policy']['id'])

    @decorators.idempotent_id('2a81795c-2a35-11e9-9d86-acde48001122')
    def test_regular_client_shares_to_another_regular_client(self):
        # owned by self.admin_client
        sg = self._create_security_group()
        with testtools.ExpectedException(exceptions.NotFound):
            self.client.show_security_group(sg['id'])
        rbac_policy = self.admin_client.create_rbac_policy(
            object_type='security_group', object_id=sg['id'],
            action='access_as_shared',
            target_tenant=self.client.project_id)['rbac_policy']
        self.client.show_security_group(sg['id'])

        self.assertIn(rbac_policy,
                      self.admin_client.list_rbac_policies()['rbac_policies'])
        # ensure that 'client2' can't see the rbac-policy sharing the
        # sg to it because the rbac-policy belongs to 'client'
        self.assertNotIn(rbac_policy['id'], [p['id'] for p in
                         self.client2.list_rbac_policies()['rbac_policies']])

    @decorators.idempotent_id('2a9fd480-2a35-11e9-9cb6-acde48001122')
    def test_filter_fields(self):
        sg = self._create_security_group()
        self.admin_client.create_rbac_policy(
            object_type='security_group', object_id=sg['id'],
            action='access_as_shared', target_tenant=self.client2.project_id)
        field_args = (('id',), ('id', 'action'), ('object_type', 'object_id'),
                      ('project_id', 'target_tenant'))
        for fields in field_args:
            res = self.admin_client.list_rbac_policies(fields=fields)
            self.assertEqual(set(fields), set(res['rbac_policies'][0].keys()))

    @decorators.idempotent_id('2abf8f9e-2a35-11e9-85f7-acde48001122')
    def test_rbac_policy_show(self):
        res = self._make_admin_sg_shared_to_project_id(
            self.client.project_id)
        p1 = res['rbac_policy']
        p2 = self.admin_client.create_rbac_policy(
            object_type='security_group',
            object_id=res['security_group']['id'],
            action='access_as_shared',
            target_tenant='*')['rbac_policy']

        self.assertEqual(
            p1, self.admin_client.show_rbac_policy(p1['id'])['rbac_policy'])
        self.assertEqual(
            p2, self.admin_client.show_rbac_policy(p2['id'])['rbac_policy'])

    @decorators.idempotent_id('2adf6bd7-2a35-11e9-9c62-acde48001122')
    def test_filter_rbac_policies(self):
        sg = self._create_security_group()
        rbac_pol1 = self.admin_client.create_rbac_policy(
            object_type='security_group', object_id=sg['id'],
            action='access_as_shared',
            target_tenant=self.client2.project_id)['rbac_policy']
        rbac_pol2 = self.admin_client.create_rbac_policy(
            object_type='security_group', object_id=sg['id'],
            action='access_as_shared',
            target_tenant=self.admin_client.project_id)['rbac_policy']
        res1 = self.admin_client.list_rbac_policies(id=rbac_pol1['id'])[
            'rbac_policies']
        res2 = self.admin_client.list_rbac_policies(id=rbac_pol2['id'])[
            'rbac_policies']
        self.assertEqual(1, len(res1))
        self.assertEqual(1, len(res2))
        self.assertEqual(rbac_pol1['id'], res1[0]['id'])
        self.assertEqual(rbac_pol2['id'], res2[0]['id'])

    @decorators.idempotent_id('2aff3900-2a35-11e9-96b3-acde48001122')
    def test_regular_client_blocked_from_sharing_anothers_policy(self):
        sg = self._make_admin_sg_shared_to_project_id(
            self.client.project_id)['security_group']
        with testtools.ExpectedException(exceptions.BadRequest):
            self.client.create_rbac_policy(
                object_type='security_group', object_id=sg['id'],
                action='access_as_shared',
                target_tenant=self.client2.project_id)

        # make sure the rbac-policy is invisible to the tenant for which it's
        # being shared
        self.assertFalse(self.client.list_rbac_policies()['rbac_policies'])


class SecGroupSearchCriteriaTest(base.BaseSearchCriteriaTest):

    required_extensions = ['security-group']
    resource = 'security-group'

    @classmethod
    def resource_setup(cls):
        super(SecGroupSearchCriteriaTest, cls).resource_setup()
        cls.security_group = cls.create_security_group()
        for name in cls.resource_names:
            cls.create_security_group(name=name)

    @decorators.idempotent_id('0064aa80-8a29-442d-a8de-9101af8210fa')
    def test_list_sorts_by_name_asc(self):
        self._test_list_sorts_asc()

    @decorators.idempotent_id('35e86832-53cd-4e63-97ec-31a2413da591')
    def test_list_sorts_by_name_desc(self):
        self._test_list_sorts_desc()

    @decorators.idempotent_id('b9654cdc-80bc-43f8-844e-dfe88fd2f125')
    def test_list_pagination(self):
        self._test_list_pagination()

    @decorators.idempotent_id('5c78bd57-e6e9-4e71-a05c-9c4082a3f139')
    def test_list_no_pagination_limit_0(self):
        self._test_list_no_pagination_limit_0()


class SecGroupNormalizedCidrTest(base.BaseNetworkTest):

    required_extensions = ['security-group', 'security-groups-normalized-cidr']

    @decorators.idempotent_id('f87bb108-205c-4f06-a378-81a26f71b829')
    def test_normalized_cidr_in_rule(self):
        security_group = self.create_security_group()
        rule = self.create_security_group_rule(
            security_group=security_group,
            direction=constants.INGRESS_DIRECTION,
            remote_ip_prefix='10.0.0.34/24')
        self.assertEqual('10.0.0.0/24', rule['normalized_cidr'])
        self.assertEqual('10.0.0.34/24', rule['remote_ip_prefix'])

        rule = self.create_security_group_rule(
            security_group=security_group,
            remote_group_id=security_group['id'],
            direction=constants.INGRESS_DIRECTION)
        self.assertIsNone(rule['normalized_cidr'])
        self.assertIsNone(rule['remote_ip_prefix'])



# Copyright (c) 2017 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from tempest.common import utils
from tempest.common import waiters
from tempest.lib.common import ssh
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base
from neutron_tempest_plugin.scenario import constants
from neutron_tempest_plugin.services.bgp import bgp_client


CONF = config.CONF


class BgpClientMixin(object):
    @classmethod
    def resource_setup(cls):
        super(BgpClientMixin, cls).resource_setup()
        if bgp_client is None:
            msg = "No BGP service client is available"
            raise cls.skipException(msg)
        manager = cls.os_admin
        cls.bgp_client = bgp_client.BgpSpeakerClientJSON(
            manager.auth_provider,
            CONF.network.catalog_type,
            CONF.network.region or CONF.identity.region,
            endpoint_type=CONF.network.endpoint_type,
            build_interval=CONF.network.build_interval,
            build_timeout=CONF.network.build_timeout,
            **manager.default_params)

    def create_bgp_speaker(self, **kwargs):
        bgp_speaker = self.bgp_client.create_bgp_speaker(post_data={
            'bgp_speaker': kwargs,
        })['bgp_speaker']
        self.addCleanup(self.bgp_client.delete_bgp_speaker, bgp_speaker['id'])
        return bgp_speaker

    def create_bgp_peer(self, **kwargs):
        bgp_peer = self.bgp_client.create_bgp_peer(post_data={
            'bgp_peer': kwargs,
        })['bgp_peer']
        self.addCleanup(self.bgp_client.delete_bgp_peer, bgp_peer['id'])
        return bgp_peer

    def add_bgp_peer_with_id(self, bgp_speaker_id, bgp_peer_id):
        self.bgp_client.add_bgp_peer_with_id(bgp_speaker_id, bgp_peer_id)


class Bgp(BgpClientMixin, base.BaseTempestTestCase):
    """Test the following topology

    .. code-block:: HTML

              +-------------------+
              | public            |
              | network           |
              |                   |
              +-+---------------+-+
                |               |
                |               |
        +-------+-+           +-+-------+
        | LEFT    |           | RIGHT   |
        | router  | <--BGP--> | router  |
        |         |           |         |
        +----+----+           +----+----+
             |                     |
        +----+----+           +----+----+
        | LEFT    |           | RIGHT   |
        | network |           | network |
        |         |           |         |
        +---------+           +---------+
    """

    credentials = ['primary', 'admin']

    @classmethod
    @utils.requires_ext(extension="bgp-speaker-router-insertion",
                        service="network")
    def resource_setup(cls):
        super(Bgp, cls).resource_setup()

        # common
        cls.keypair = cls.create_keypair()
        cls.secgroup = cls.os_primary.network_client.create_security_group(
            name=data_utils.rand_name('secgroup-'))['security_group']
        cls.security_groups.append(cls.secgroup)
        cls.create_loginable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.create_pingable_secgroup_rule(secgroup_id=cls.secgroup['id'])

        # LEFT
        cls.router = cls.create_router(
            data_utils.rand_name('left-router'),
            admin_state_up=True,
            external_network_id=CONF.network.public_network_id)
        cls.network = cls.create_network(network_name='left-network')
        cls.subnet = cls.create_subnet(cls.network,
                                       name='left-subnet')
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

        # RIGHT
        cls._right_network, cls._right_subnet, cls._right_router = \
            cls._create_right_network()

    @classmethod
    def _create_right_network(cls):
        # NOTE(yamamoto): Disable SNAT to workaround a bug
        # https://midonet.atlassian.net/browse/MNA-1114
        router = cls.create_admin_router(
            data_utils.rand_name('right-router'),
            admin_state_up=True,
            external_network_id=CONF.network.public_network_id,
            enable_snat=False,
            project_id=cls.os_primary.network_client.project_id)
        network = cls.create_network(network_name='right-network')
        subnet = cls.create_subnet(
            network,
            cidr=netaddr.IPNetwork('10.10.0.0/24'),
            name='right-subnet')
        cls.create_router_interface(router['id'], subnet['id'])
        return network, subnet, router

    def _create_server(self, create_floating_ip=True, network=None):
        if network is None:
            network = self.network
        port = self.create_port(network, security_groups=[self.secgroup['id']])
        if create_floating_ip:
            fip = self.create_and_associate_floatingip(port['id'])
        else:
            fip = None
        server = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=self.keypair['name'],
            networks=[{'port': port['id']}])['server']
        waiters.wait_for_server_status(self.os_primary.servers_client,
                                       server['id'],
                                       constants.SERVER_STATUS_ACTIVE)
        return {'port': port, 'fip': fip, 'server': server}

    def _find_ipv4_subnet(self, network_id):
        subnets = self.os_admin.network_client.list_subnets(
            network_id=network_id)['subnets']
        for subnet in subnets:
            if subnet['ip_version'] == 4:
                return subnet['id']
        msg = "No suitable subnets on public network"
        raise self.skipException(msg)

    def _get_external_ip(self, router, subnet_id):
        for ip in router['external_gateway_info']['external_fixed_ips']:
            if ip['subnet_id'] == subnet_id:
                return ip['ip_address']
        return None

    def _setup_bgp(self):
        network_id = CONF.network.public_network_id
        subnet_id = self._find_ipv4_subnet(network_id)
        sites = [
            dict(name="left", network=self.network, subnet=self.subnet,
                 router=self.router, local_as=64512),
            dict(name="right", network=self._right_network,
                 subnet=self._right_subnet, router=self._right_router,
                 local_as=64513),
        ]
        psk = data_utils.rand_name('mysecret')
        for i in range(0, 2):
            site = sites[i]
            router = site['router']
            site['bgp_speaker'] = self.create_bgp_speaker(
                name=data_utils.rand_name('%s-bgp-speaker' % site['name']),
                local_as=site['local_as'],
                ip_version=4,
                logical_router=router['id'])
            site['external_v4_ip'] = self._get_external_ip(router, subnet_id)
        for i in range(0, 2):
            site = sites[i]
            bgp_speaker_id = site['bgp_speaker']['id']
            peer = sites[1 - i]
            peer_ip = peer['external_v4_ip']
            peer_as = peer['local_as']
            bgp_peer = self.create_bgp_peer(
                name=data_utils.rand_name('%s-bgp-peer' % site['name']),
                peer_ip=peer_ip,
                remote_as=peer_as,
                auth_type='md5',
                password=psk)
            self.add_bgp_peer_with_id(bgp_speaker_id, bgp_peer['id'])

    @decorators.idempotent_id('c1208ce2-c55f-4424-9035-25de83161d6f')
    def test_bgp(self):
        # RIGHT
        right_server = self._create_server(
            network=self._right_network,
            create_floating_ip=False)

        # LEFT
        left_server = self._create_server()
        ssh_client = ssh.Client(left_server['fip']['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                pkey=self.keypair['private_key'],
                                ssh_key_type=CONF.validation.ssh_key_type)

        # check LEFT -> RIGHT connectivity via BGP advertised routes
        self.check_remote_connectivity(
            ssh_client,
            right_server['port']['fixed_ips'][0]['ip_address'],
            should_succeed=False)
        self._setup_bgp()
        self.check_remote_connectivity(
            ssh_client,
            right_server['port']['fixed_ips'][0]['ip_address'])



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from neutron_lib import constants
from neutron_lib.utils import test
from oslo_log import log as logging
from tempest.common import utils as tutils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
import testtools

from neutron_tempest_plugin.common import ip
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base


LOG = logging.getLogger(__name__)
CONF = config.CONF


ServerWithTrunkPort = collections.namedtuple(
    'ServerWithTrunkPort',
    ['port', 'subport', 'trunk', 'floating_ip', 'server',
     'ssh_client'])


class TrunkTest(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    @classmethod
    @tutils.requires_ext(extension="trunk", service="network")
    def resource_setup(cls):
        super(TrunkTest, cls).resource_setup()
        # setup basic topology for servers we can log into
        cls.rand_name = data_utils.rand_name(
            cls.__name__.rsplit('.', 1)[-1])
        cls.network = cls.create_network(name=cls.rand_name)
        cls.subnet = cls.create_subnet(network=cls.network,
                                       name=cls.rand_name)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.keypair = cls.create_keypair(name=cls.rand_name)

    def setUp(self):
        super(TrunkTest, self).setUp()
        self.security_group = self.create_security_group(name=self.rand_name)
        self.create_loginable_secgroup_rule(self.security_group['id'])

    def _create_server_with_network(self, network, use_advanced_image=False):
        port = self._create_server_port(network=network)
        floating_ip = self.create_floatingip(port=port)
        ssh_client = self._create_ssh_client(
            floating_ip=floating_ip, use_advanced_image=use_advanced_image)
        server = self._create_server(port=port,
                                     use_advanced_image=use_advanced_image)
        return ServerWithTrunkPort(port=port, subport=None, trunk=None,
                                   floating_ip=floating_ip, server=server,
                                   ssh_client=ssh_client)

    def _create_server_with_trunk_port(self, subport_network=None,
                                       segmentation_id=None,
                                       use_advanced_image=False):
        port = self._create_server_port()
        floating_ip = self.create_floatingip(port=port)
        ssh_client = self._create_ssh_client(
            floating_ip=floating_ip, use_advanced_image=use_advanced_image)

        subport = None
        subports = None
        if subport_network:
            subport = self._create_server_port(
                network=subport_network, mac_address=port['mac_address'])
            subports = [{'port_id': subport['id'],
                         'segmentation_type': 'vlan',
                         'segmentation_id': segmentation_id}]
        trunk = self.create_trunk(port=port, subports=subports)

        server = self._create_server(port=port,
                                     use_advanced_image=use_advanced_image)
        return ServerWithTrunkPort(port=port, subport=subport, trunk=trunk,
                                   floating_ip=floating_ip, server=server,
                                   ssh_client=ssh_client)

    def _create_advanced_servers_with_trunk_port(self, num_servers=1,
                                                 subport_network=None,
                                                 segmentation_id=None,
                                                 vlan_subnet=None,
                                                 use_advanced_image=False):
        server_list = []
        for _ in range(0, num_servers):
            vm = self._create_server_with_trunk_port(
                subport_network,
                segmentation_id,
                use_advanced_image)
            server_list.append(vm)
            self._configure_vlan_subport(
                vm=vm,
                vlan_tag=segmentation_id,
                vlan_subnet=vlan_subnet)

        return server_list

    def _check_servers_remote_connectivity(self, vms=None,
                                           should_succeed=True):
        self.check_remote_connectivity(
            vms[0].ssh_client,
            vms[1].subport['fixed_ips'][0]['ip_address'],
            should_succeed=should_succeed)

    def _create_server_port(self, network=None, **params):
        network = network or self.network
        return self.create_port(network=network, name=self.rand_name,
                                security_groups=[self.security_group['id']],
                                **params)

    def _create_server(self, port, use_advanced_image=False, **params):
        if use_advanced_image:
            flavor_ref = CONF.neutron_plugin_options.advanced_image_flavor_ref
            image_ref = CONF.neutron_plugin_options.advanced_image_ref
        else:
            flavor_ref = CONF.compute.flavor_ref
            image_ref = CONF.compute.image_ref
        return self.create_server(flavor_ref=flavor_ref,
                                  image_ref=image_ref,
                                  key_name=self.keypair['name'],
                                  networks=[{'port': port['id']}],
                                  **params)['server']

    def _show_port(self, port, update=False):
        observed = self.client.show_port(port['id'])['port']
        if update:
            port.update(observed)
        return observed

    def _show_trunk(self, trunk, update=False):
        observed = self.client.show_trunk(trunk['id'])['trunk']
        if update:
            trunk.update(observed)
        return observed

    def _is_trunk_status(self, trunk, status, update=False):
        return self._show_trunk(trunk, update)['status'] == status

    def _is_port_status(self, port, status, update=False):
        return self._show_port(port, update)['status'] == status

    def _wait_for_port(self, port, status=constants.ACTIVE):
        utils.wait_until_true(
            lambda: self._is_port_status(port, status),
            exception=RuntimeError(
                "Timed out waiting for port {!r} to transition to get "
                "status {!r}.".format(port['id'], status)))

    def _wait_for_trunk(self, trunk, status=constants.ACTIVE):
        utils.wait_until_true(
            lambda: self._is_trunk_status(trunk, status),
            exception=RuntimeError(
                "Timed out waiting for trunk {!r} to transition to get "
                "status {!r}.".format(trunk['id'], status)))

    def _create_ssh_client(self, floating_ip, use_advanced_image=False):
        if use_advanced_image:
            username = CONF.neutron_plugin_options.advanced_image_ssh_user
        else:
            username = CONF.validation.image_ssh_user
        return ssh.Client(host=floating_ip['floating_ip_address'],
                          username=username,
                          pkey=self.keypair['private_key'])

    def _assert_has_ssh_connectivity(self, ssh_client):
        ssh_client.exec_command("true")

    def _configure_vlan_subport(self, vm, vlan_tag, vlan_subnet):
        self.wait_for_server_active(server=vm.server)
        self.wait_for_guest_os_ready(vm.server)
        self._wait_for_trunk(trunk=vm.trunk)
        self._wait_for_port(port=vm.port)
        self._wait_for_port(port=vm.subport)
        self.check_connectivity(
            host=vm.floating_ip['floating_ip_address'],
            ssh_client=vm.ssh_client,
            servers=[vm.server])

        ip_command = ip.IPCommand(ssh_client=vm.ssh_client)
        for address in ip_command.list_addresses(port=vm.port):
            port_iface = address.device.name
            break
        else:
            self.fail("Parent port fixed IP not found on server.")

        subport_iface = ip_command.configure_vlan_subport(
            port=vm.port, subport=vm.subport, vlan_tag=vlan_tag,
            subnets=[vlan_subnet])
        for address in ip_command.list_addresses(port=vm.subport):
            self.assertEqual(subport_iface, address.device.name)
            self.assertEqual(port_iface, address.device.parent)
            break
        else:
            self.fail("Sub-port fixed IP not found on server.")

    @decorators.idempotent_id('bb13fe28-f152-4000-8131-37890a40c79e')
    def test_trunk_subport_lifecycle(self):
        """Test trunk creation and subport transition to ACTIVE status.

        This is a basic test for the trunk extension to ensure that we
        can create a trunk, attach it to a server, add/remove subports,
        while ensuring the status transitions as appropriate.

        This test does not assert any dataplane behavior for the subports.
        It's just a high-level check to ensure the agents claim to have
        wired the port correctly and that the trunk port itself maintains
        connectivity.
        """
        vm1 = self._create_server_with_trunk_port()
        vm2 = self._create_server_with_trunk_port()
        for vm in (vm1, vm2):
            self.wait_for_server_active(server=vm.server)
            self.wait_for_guest_os_ready(vm.server)
            self._wait_for_trunk(vm.trunk)
            self._assert_has_ssh_connectivity(vm.ssh_client)

        # create a few more networks and ports for subports
        # check limit of networks per project
        segment_ids = range(
            3, 3 + CONF.neutron_plugin_options.max_networks_per_project)
        tagged_networks = [self.create_network() for _ in segment_ids]
        tagged_ports = [self.create_port(network=network)
                        for network in tagged_networks]
        subports = [{'port_id': tagged_ports[i]['id'],
                     'segmentation_type': 'vlan',
                     'segmentation_id': segment_id}
                    for i, segment_id in enumerate(segment_ids)]

        # add all subports to server1
        self.client.add_subports(vm1.trunk['id'], subports)
        self._wait_for_trunk(vm1.trunk)
        for port in tagged_ports:
            self._wait_for_port(port)

        # ensure main data-plane wasn't interrupted
        self._assert_has_ssh_connectivity(vm1.ssh_client)

        # move subports over to other server
        self.client.remove_subports(vm1.trunk['id'], subports)
        # ensure all subports go down
        for port in tagged_ports:
            self._wait_for_port(port, status=constants.DOWN)

        self.client.add_subports(vm2.trunk['id'], subports)

        # wait for both trunks to go back to ACTIVE
        for vm in [vm1, vm2]:
            self._wait_for_trunk(vm.trunk)

        # ensure subports come up on other trunk
        for port in tagged_ports:
            self._wait_for_port(port)

        # final connectivity check
        for vm in [vm1, vm2]:
            self._wait_for_trunk(vm.trunk)
            self._assert_has_ssh_connectivity(vm1.ssh_client)

    @test.unstable_test("bug 2033887 / bug 2024160")
    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @testtools.skipUnless(
         (CONF.neutron_plugin_options.reboots_in_test > 0),
         "Number of reboots > 0 is reqired for this test")
    @decorators.idempotent_id('a8a02c9b-b453-49b5-89a2-cce7da6680fb')
    def test_subport_connectivity_soft_reboot(self):
        vlan_tag = 10
        vlan_network = self.create_network()
        vlan_subnet = self.create_subnet(network=vlan_network, gateway=None)
        use_advanced_image = (
            not CONF.neutron_plugin_options.default_image_is_advanced)

        # allow intra-security-group traffic
        sg_rule = self.create_pingable_secgroup_rule(self.security_group['id'])
        self.addCleanup(
            self.os_primary.network_client.delete_security_group_rule,
            sg_rule['id'])

        vms = self._create_advanced_servers_with_trunk_port(
            num_servers=2,
            subport_network=vlan_network,
            segmentation_id=vlan_tag,
            vlan_subnet=vlan_subnet,
            use_advanced_image=use_advanced_image)
        # check remote connectivity true before reboots
        self._check_servers_remote_connectivity(vms=vms)
        client = self.os_tempest.compute.ServersClient()
        for _ in range(CONF.neutron_plugin_options.reboots_in_test):
            client.reboot_server(vms[1].server['id'],
                                 **{'type': 'SOFT'})
            self.wait_for_server_active(vms[1].server)
            self._configure_vlan_subport(vm=vms[1],
                                         vlan_tag=vlan_tag,
                                         vlan_subnet=vlan_subnet)
            self._check_servers_remote_connectivity(vms=vms)

    @test.unstable_test("bug 1897796")
    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @decorators.idempotent_id('a8a02c9b-b453-49b5-89a2-cce7da66bbcb')
    def test_subport_connectivity(self):
        vlan_tag = 10
        vlan_network = self.create_network()
        vlan_subnet = self.create_subnet(network=vlan_network, gateway=None)
        use_advanced_image = (
            not CONF.neutron_plugin_options.default_image_is_advanced)
        vms = self._create_advanced_servers_with_trunk_port(
            num_servers=2,
            subport_network=vlan_network,
            segmentation_id=vlan_tag,
            vlan_subnet=vlan_subnet,
            use_advanced_image=use_advanced_image)
        # Ping from server1 to server2 via VLAN interface should fail because
        # we haven't allowed ICMP
        self._check_servers_remote_connectivity(vms=vms,
                                                should_succeed=False)
        # allow intra-security-group traffic
        sg_rule = self.create_pingable_secgroup_rule(self.security_group['id'])
        self.addCleanup(
                self.os_primary.network_client.delete_security_group_rule,
                sg_rule['id'])
        self._check_servers_remote_connectivity(vms=vms)

    @testtools.skipUnless(CONF.compute_feature_enabled.cold_migration,
                          'Cold migration is not available.')
    @testtools.skipUnless(CONF.compute.min_compute_nodes > 1,
                          'Less than 2 compute nodes, skipping multinode '
                          'tests.')
    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @decorators.attr(type='slow')
    @decorators.idempotent_id('ecd7de30-1c90-4280-b97c-1bed776d5d07')
    def test_trunk_vm_migration(self):
        '''Test connectivity after migration of the server with trunk

        A successfully migrated server shows a VERIFY_RESIZE status that
        requires confirmation. Need to reconfigure VLAN interface on server
        side after migration is finished as the configuration doesn't survive
        the reboot.
        '''
        vlan_tag = 10
        vlan_network = self.create_network()
        vlan_subnet = self.create_subnet(vlan_network)
        sg_rule = self.create_pingable_secgroup_rule(self.security_group['id'])
        self.addCleanup(
                self.os_primary.network_client.delete_security_group_rule,
                sg_rule['id'])

        use_advanced_image = (
            not CONF.neutron_plugin_options.default_image_is_advanced)
        servers = {}
        for role in ['migrate', 'connection_test']:
            servers[role] = self._create_server_with_trunk_port(
                                subport_network=vlan_network,
                                segmentation_id=vlan_tag,
                                use_advanced_image=use_advanced_image)
        for role in ['migrate', 'connection_test']:
            self.wait_for_server_active(servers[role].server)
            self.wait_for_guest_os_ready(servers[role].server)
            self._configure_vlan_subport(vm=servers[role],
                                         vlan_tag=vlan_tag,
                                         vlan_subnet=vlan_subnet)

        self.check_remote_connectivity(
                servers['connection_test'].ssh_client,
                servers['migrate'].subport['fixed_ips'][0]['ip_address'])

        client = self.os_admin.compute.ServersClient()
        client.migrate_server(servers['migrate'].server['id'])
        self.wait_for_server_status(servers['migrate'].server,
                                    'VERIFY_RESIZE')
        client.confirm_resize_server(servers['migrate'].server['id'])
        self._configure_vlan_subport(vm=servers['migrate'],
                                     vlan_tag=vlan_tag,
                                     vlan_subnet=vlan_subnet)

        self.check_remote_connectivity(
                servers['connection_test'].ssh_client,
                servers['migrate'].subport['fixed_ips'][0]['ip_address'])

    @testtools.skipUnless(
        (CONF.neutron_plugin_options.advanced_image_ref or
         CONF.neutron_plugin_options.default_image_is_advanced),
        "Advanced image is required to run this test.")
    @testtools.skipUnless(
        CONF.neutron_plugin_options.q_agent == "linuxbridge",
        "Linux bridge agent is required to run this test.")
    @decorators.idempotent_id('d61cbdf6-1896-491c-b4b4-871caf7fbffe')
    def test_parent_port_connectivity_after_trunk_deleted_lb(self):
        vlan_tag = 10
        vlan_network = self.create_network()
        vlan_subnet = self.create_subnet(vlan_network)
        self.create_router_interface(self.router['id'], vlan_subnet['id'])

        use_advanced_image = (
            not CONF.neutron_plugin_options.default_image_is_advanced)

        # Create servers
        trunk_network_server = self._create_server_with_trunk_port(
            subport_network=vlan_network,
            segmentation_id=vlan_tag,
            use_advanced_image=use_advanced_image)
        normal_network_server = self._create_server_with_network(self.network)
        vlan_network_server = self._create_server_with_network(vlan_network)
        vms = [normal_network_server, vlan_network_server]

        self._configure_vlan_subport(vm=trunk_network_server,
                                     vlan_tag=vlan_tag,
                                     vlan_subnet=vlan_subnet)
        for vm in vms:
            self.wait_for_server_active(vm.server)
            self.wait_for_guest_os_ready(vm.server)

        # allow ICMP traffic
        sg_rule = self.create_pingable_secgroup_rule(self.security_group['id'])
        self.addCleanup(
                self.os_primary.network_client.delete_security_group_rule,
                sg_rule['id'])

        # Ping from trunk_network_server to normal_network_server
        # via parent port
        self.check_remote_connectivity(
            trunk_network_server.ssh_client,
            normal_network_server.port['fixed_ips'][0]['ip_address'],
            should_succeed=True,
            servers=vms)

        # Ping from trunk_network_server to vlan_network_server via VLAN
        # interface should success
        self.check_remote_connectivity(
            trunk_network_server.ssh_client,
            vlan_network_server.port['fixed_ips'][0]['ip_address'],
            should_succeed=True,
            servers=vms)

        # Delete the trunk
        self.delete_trunk(
            trunk_network_server.trunk,
            detach_parent_port=False)
        LOG.debug("Trunk %s is deleted.",
                  trunk_network_server.trunk['id'])

        # Ping from trunk_network_server to normal_network_server
        # via parent port success after trunk deleted
        self.check_remote_connectivity(
            trunk_network_server.ssh_client,
            normal_network_server.port['fixed_ips'][0]['ip_address'],
            should_succeed=True,
            servers=vms)

        # Ping from trunk_network_server to vlan_network_server via VLAN
        # interface should fail after trunk deleted
        self.check_remote_connectivity(
            trunk_network_server.ssh_client,
            vlan_network_server.port['fixed_ips'][0]['ip_address'],
            should_succeed=False)



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest import config

from neutron_tempest_plugin.sfc.services import flowclassifier_client

CONF = config.CONF


class FlowClassifierClientMixin(object):

    @classmethod
    def resource_setup(cls):
        super(FlowClassifierClientMixin, cls).resource_setup()
        manager = cls.os_admin
        cls.flowclassifier_client = (
            flowclassifier_client.FlowClassifierClient(
                manager.auth_provider,
                CONF.network.catalog_type,
                CONF.network.region or CONF.identity.region,
                endpoint_type=CONF.network.endpoint_type,
                build_interval=CONF.network.build_interval,
                build_timeout=CONF.network.build_timeout,
                **manager.default_params
            )
        )

    @classmethod
    def create_flowclassifier(cls, **kwargs):
        body = cls.flowclassifier_client.create_flowclassifier(
            **kwargs)
        fc = body['flow_classifier']
        return fc



# Copyright (c) 2019 AT&T
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager
from oslo_log import log
import testtools

from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils.linux import remote_client
from tempest.lib.common.utils import test_utils

from tempest.lib import decorators

from neutron_tempest_plugin.tap_as_a_service.scenario import manager


CONF = config.CONF
LOG = log.getLogger(__name__)


class TestTaaSTrafficScenarios(manager.BaseTaasScenarioTests):

    @classmethod
    @utils.requires_ext(extension='taas', service='network')
    @utils.requires_ext(extension='security-group', service='network')
    @utils.requires_ext(extension='router', service='network')
    def skip_checks(cls):
        super(TestTaaSTrafficScenarios, cls).skip_checks()

    @classmethod
    def resource_setup(cls):
        super(TestTaaSTrafficScenarios, cls).resource_setup()
        cls.provider_network = None
        cls.keypair = cls.create_keypair()
        cls.secgroup = cls.create_security_group(
            name=data_utils.rand_name('secgroup'))
        cls.create_loginable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.create_pingable_secgroup_rule(secgroup_id=cls.secgroup['id'])

    @contextmanager
    def _setup_topology(self, taas=True, use_taas_cloud_image=False,
                        provider_net=False):
        """Setup topology for the test

           +------------+
           | monitor vm |
           +-----+------+
                 |
           +-----v---+
        +--+ network <--+
        |  +----^----+  |
        |       |       |
        |  +----+-+ +---+--+
        |  | vm 1 | | vm 2 |
        |  +------+ +------+
        |
        |  +--------+
        +--> router |
           +-----+--+
                 |
           +-----v------+
           | public net |
           +------------+
       """
        self.network, self.subnet, self.router = self.create_networks()
        LOG.debug('Setup topology sbunet details: %s ', self.subnet)
        if provider_net:
            if CONF.taas.provider_physical_network:
                self.provider_network = self._setup_provider_network()
            else:
                msg = "provider_physical_network not provided"
                raise self.skipException(msg)

        self.mon_port, mon_fip = self._create_server_with_floatingip(
            use_taas_cloud_image=use_taas_cloud_image,
            provider_net=provider_net)
        LOG.debug('Setup topology monitor port: %s  ###  monitor FIP: %s ',
                  self.mon_port, mon_fip)
        self.left_port, self.left_fip = self._create_server_with_floatingip(
            provider_net=provider_net)
        LOG.debug('Setup topology left port: %s  ###  left FIP: %s ',
                  self.left_port, self.left_fip)
        self.right_port, self.right_fip = self._create_server_with_floatingip(
            provider_net=provider_net)
        LOG.debug('Setup topology right port: %s  ###  right FIP: %s ',
                  self.right_port, self.right_fip)

        if taas:
            LOG.debug("Create TAAS service")
            tap_service = self.tap_services_client.create_tap_service(
                port_id=self.mon_port['id'])['tap_service']
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.client.delete_tap_service, tap_service['id'])
            tap_flow = self.tap_flows_client.create_tap_flow(
                tap_service_id=tap_service['id'], direction='BOTH',
                source_port=self.left_port['id'])['tap_flow']
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.client.delete_tap_flow, tap_flow['id'])
            tap_flow = self.tap_flows_client.create_tap_flow(
                tap_service_id=tap_service['id'], direction='BOTH',
                source_port=self.right_port['id'])['tap_flow']
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            self.client.delete_tap_flow, tap_flow['id'])

        user = CONF.validation.image_ssh_user
        if use_taas_cloud_image:
            user = CONF.neutron_plugin_options.advanced_image_ssh_user

        self.monitor_client = remote_client.RemoteClient(
            mon_fip['floating_ip_address'], user,
            pkey=self.keypair['private_key'],
            ssh_key_type=CONF.validation.ssh_key_type)
        self.monitor_client.validate_authentication()
        self.left_client = remote_client.RemoteClient(
            self.left_fip['floating_ip_address'],
            CONF.validation.image_ssh_user,
            pkey=self.keypair['private_key'],
            ssh_key_type=CONF.validation.ssh_key_type)
        self.left_client.validate_authentication()
        self.right_client = remote_client.RemoteClient(
            self.right_fip['floating_ip_address'],
            CONF.validation.image_ssh_user,
            pkey=self.keypair['private_key'],
            ssh_key_type=CONF.validation.ssh_key_type)
        self.right_client.validate_authentication()
        yield

    def _check_icmp_traffic(self):
        log_location = "/tmp/tcpdumplog"

        right_ip = self.right_port['fixed_ips'][0]['ip_address']
        left_ip = self.left_port['fixed_ips'][0]['ip_address']

        # Run tcpdump in background
        self._run_in_background(self.monitor_client,
                                "sudo tcpdump -n -nn > %s" % log_location)

        # Ensure tcpdump is up and running
        psax = self.monitor_client.exec_command("ps -ax")
        self.assertTrue("tcpdump" in psax)

        # Run traffic from left_vm to right_vm
        LOG.debug('Check ICMP traffic: ping %s ', right_ip)
        # self.left_client.exec_command(
        #     "ping -c 50 %s" % self.right_fip['floating_ip_address'])
        self.check_remote_connectivity(self.left_client, right_ip,
                                       ping_count=50)

        # Collect tcpdump results
        output = self.monitor_client.exec_command("cat %s" % log_location)
        self.assertLess(0, len(output))

        looking_for = ["IP %s > %s: ICMP echo request" % (left_ip, right_ip),
                       "IP %s > %s: ICMP echo reply" % (right_ip, left_ip)]

        results = []
        for tcpdump_line in looking_for:
            results.append(tcpdump_line in output)

        return all(results)

    def _test_taas_connectivity(self, use_provider_net=False):
        """Ensure TAAS doesn't break connectivity

        This test creates TAAS service between two servers and checks that
        it doesn't break basic connectivity between them.
        """
        # Check uninterrupted traffic between VMs
        with self._setup_topology(provider_net=use_provider_net):
            # Left to right
            self.check_remote_connectivity(
                self.left_client,
                self.right_port['fixed_ips'][0]['ip_address'])

            # Right to left
            self.check_remote_connectivity(
                self.right_client,
                self.left_port['fixed_ips'][0]['ip_address'])

            # TAAS vm to right
            self.check_remote_connectivity(
                self.monitor_client,
                self.right_port['fixed_ips'][0]['ip_address'])

            # TAAS vm to left
            self.check_remote_connectivity(
                self.monitor_client,
                self.left_port['fixed_ips'][0]['ip_address'])

    @decorators.idempotent_id('ff414b7d-e81c-47f2-b6c8-53bc2f1e9b00')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_provider_network_connectivity(self):
        self._test_taas_connectivity(use_provider_net=True)

    @decorators.idempotent_id('e3c52e91-7abf-4dfd-8687-f7c071cdd333')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_network_connectivity(self):
        self._test_taas_connectivity(use_provider_net=False)

    @decorators.idempotent_id('fcb15ca3-ef61-11e9-9792-f45c89c47e11')
    @testtools.skipUnless(CONF.neutron_plugin_options.advanced_image_ref,
                          'Cloud image not found.')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_forwarded_traffic_positive(self):
        """Check that TAAS forwards traffic as expected"""

        with self._setup_topology(use_taas_cloud_image=True):
            # Check that traffic was forwarded to TAAS service
            self.assertTrue(self._check_icmp_traffic())

    @decorators.idempotent_id('6c54d9c5-075a-4a1f-bbe6-12c3c9abf1e2')
    @testtools.skipUnless(CONF.neutron_plugin_options.advanced_image_ref,
                          'Cloud image not found.')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_forwarded_traffic_negative(self):
        """Check that TAAS doesn't forward traffic"""

        with self._setup_topology(taas=False, use_taas_cloud_image=True):
            # Check that traffic was NOT forwarded to TAAS service
            self.assertFalse(self._check_icmp_traffic())

    @decorators.idempotent_id('fcb15ca3-ef61-11e9-9792-f45c89c47e12')
    @testtools.skipUnless(CONF.neutron_plugin_options.advanced_image_ref,
                          'Cloud image not found.')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_forwarded_traffic_provider_net_positive(self):
        """Check that TAAS forwards traffic as expected in provider network"""

        with self._setup_topology(use_taas_cloud_image=True,
                                  provider_net=True):
            # Check that traffic was forwarded to TAAS service
            self.assertTrue(self._check_icmp_traffic())

    @decorators.idempotent_id('6c54d9c5-075a-4a1f-bbe6-12c3c9abf1e3')
    @testtools.skipUnless(CONF.neutron_plugin_options.advanced_image_ref,
                          'Cloud image not found.')
    @decorators.attr(type='slow')
    @utils.services('compute', 'network')
    def test_taas_forwarded_traffic_provider_net_negative(self):
        """Check that TAAS doesn't forward traffic in provider network"""

        with self._setup_topology(taas=False, use_taas_cloud_image=True,
                                  provider_net=True):
            # Check that traffic was NOT forwarded to TAAS service
            self.assertFalse(self._check_icmp_traffic())



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import exceptions

from neutron_tempest_plugin.api import base


class BaseRouterTest(base.BaseAdminNetworkTest):
    # NOTE(salv-orlando): This class inherits from BaseAdminNetworkTest
    # as some router operations, such as enabling or disabling SNAT
    # require admin credentials by default

    def _cleanup_router(self, router, client=None):
        try:
            self.delete_router(router, client)
            self.routers.remove(router)
        except exceptions.NotFound:
            pass

    def _create_router(self, name, admin_state_up=False,
                       external_network_id=None, enable_snat=None):
        # associate a cleanup with created routers to avoid quota limits
        router = self.create_router(name, admin_state_up,
                                    external_network_id, enable_snat)
        self.addCleanup(self._cleanup_router, router)
        return router

    def _create_admin_router(self, *args, **kwargs):
        router = self.create_admin_router(*args, **kwargs)
        self.addCleanup(
            self._cleanup_router, router, self.os_admin.network_client)
        return router

    def _delete_router(self, router_id, network_client=None):
        client = network_client or self.client
        client.delete_router(router_id)
        # Asserting that the router is not found in the list
        # after deletion
        list_body = self.client.list_routers()
        routers_list = list()
        for router in list_body['routers']:
            routers_list.append(router['id'])
        self.assertNotIn(router_id, routers_list)



# Copyright 2012 OpenStack Foundation
# Copyright 2016 Hewlett Packard Enterprise Development Company LP
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config
from neutron_tempest_plugin.vpnaas.services import clients_vpnaas

CONF = config.CONF


class BaseNetworkTest(base.BaseNetworkTest):
    @classmethod
    def resource_setup(cls):
        super(BaseNetworkTest, cls).resource_setup()
        cls.vpnservices = []
        cls.ikepolicies = []
        cls.ipsecpolicies = []
        cls.ipsec_site_connections = []
        cls.endpoint_groups = []

    @classmethod
    def get_client_manager(cls, credential_type=None, roles=None,
                           force_new=None):
        manager = super(BaseNetworkTest, cls).get_client_manager(
            credential_type=credential_type,
            roles=roles,
            force_new=force_new)
        # Neutron uses a different clients manager than the one in the Tempest
        return clients_vpnaas.Manager(manager.credentials)

    @classmethod
    def resource_cleanup(cls):
        if CONF.service_available.neutron:
            # Clean up ipsec connections
            for ipsec_site_connection in cls.ipsec_site_connections:
                cls._try_delete_resource(
                    cls.client.delete_ipsec_site_connection,
                    ipsec_site_connection['id'])

            # Clean up ipsec endpoint group
            for endpoint_group in cls.endpoint_groups:
                cls._try_delete_resource(cls.client.delete_endpoint_group,
                                         endpoint_group['id'])

            # Clean up ipsec policies
            for ipsecpolicy in cls.ipsecpolicies:
                cls._try_delete_resource(cls.client.delete_ipsecpolicy,
                                         ipsecpolicy['id'])
            # Clean up ike policies
            for ikepolicy in cls.ikepolicies:
                cls._try_delete_resource(cls.client.delete_ikepolicy,
                                         ikepolicy['id'])
            # Clean up vpn services
            for vpnservice in cls.vpnservices:
                cls._try_delete_resource(cls.client.delete_vpnservice,
                                         vpnservice['id'])
        super(BaseNetworkTest, cls).resource_cleanup()

    @classmethod
    def create_vpnservice(cls, subnet_id, router_id, name=None):
        """Wrapper utility that returns a test vpn service."""
        if name is None:
            name = data_utils.rand_name("vpnservice-")
        body = cls.client.create_vpnservice(
            subnet_id=subnet_id, router_id=router_id, admin_state_up=True,
            name=name)
        vpnservice = body['vpnservice']
        cls.vpnservices.append(vpnservice)
        return vpnservice

    @classmethod
    def create_vpnservice_no_subnet(cls, router_id):
        """Wrapper utility that returns a test vpn service."""
        body = cls.client.create_vpnservice(
            router_id=router_id, admin_state_up=True,
            name=data_utils.rand_name("vpnservice-"))
        vpnservice = body['vpnservice']
        cls.vpnservices.append(vpnservice)
        return vpnservice

    @classmethod
    def create_ikepolicy(cls, name):
        """Wrapper utility that returns a test ike policy."""
        body = cls.client.create_ikepolicy(name=name)
        ikepolicy = body['ikepolicy']
        cls.ikepolicies.append(ikepolicy)
        return ikepolicy

    @classmethod
    def create_ipsecpolicy(cls, name):
        """Wrapper utility that returns a test ipsec policy."""
        body = cls.client.create_ipsecpolicy(name=name)
        ipsecpolicy = body['ipsecpolicy']
        cls.ipsecpolicies.append(ipsecpolicy)
        return ipsecpolicy

    @classmethod
    def create_ipsec_site_connection(cls, ikepolicy_id, ipsecpolicy_id,
                                     vpnservice_id, psk="secret",
                                     peer_address="172.24.4.233",
                                     peer_id="172.24.4.233",
                                     peer_cidrs=None,
                                     name=None):
        """Wrapper utility that returns a test vpn connection."""
        if peer_cidrs is None:
            peer_cidrs = ['1.1.1.0/24', '2.2.2.0/24']
        if name is None:
            name = data_utils.rand_name("ipsec_site_connection-")
        body = cls.client.create_ipsec_site_connection(
            psk=psk,
            initiator="bi-directional",
            ipsecpolicy_id=ipsecpolicy_id,
            admin_state_up=True,
            mtu=1500,
            ikepolicy_id=ikepolicy_id,
            vpnservice_id=vpnservice_id,
            peer_address=peer_address,
            peer_id=peer_id,
            peer_cidrs=peer_cidrs,
            name=name)
        ipsec_site_connection = body['ipsec_site_connection']
        cls.ipsec_site_connections.append(ipsec_site_connection)
        return ipsec_site_connection

    @classmethod
    def create_endpoint_group(cls, name, type, endpoints):
        """Wrapper utility that returns a test ipsec policy."""
        body = cls.client.create_endpoint_group(
            endpoints=endpoints,
            type=type,
            description='endpoint type:' + type,
            name=name)
        endpoint_group = body['endpoint_group']
        cls.endpoint_groups.append(endpoint_group)
        return endpoint_group


class BaseAdminNetworkTest(BaseNetworkTest):
    credentials = ['primary', 'admin']

    @classmethod
    def setup_clients(cls):
        super(BaseAdminNetworkTest, cls).setup_clients()
        cls.admin_client = cls.os_admin.network_client
        cls.identity_admin_client = cls.os_admin.tenants_client



# Copyright 2019 SUSE LLC
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from tempest.common import utils
from tempest.lib import decorators

from neutron_tempest_plugin.api import test_subnetpools

SUBNETPOOL_NAME = 'smoke-subnetpool'
SUBNET_NAME = 'smoke-subnet'


class SubnetPoolPrefixOpsTestMixin(object):

    def _compare_prefix_lists(self, list_expected, list_observed):
        expected_set = netaddr.IPSet(iterable=list_expected)
        observed_set = netaddr.IPSet(iterable=list_observed)

        # compact the IPSet's
        expected_set.compact()
        observed_set.compact()

        self.assertEqual(expected_set, observed_set)

    @decorators.idempotent_id('b1d56d1f-2818-44ee-b6a3-3c1327c25318')
    @utils.requires_ext(extension='subnetpool-prefix-ops', service='network')
    def test_add_remove_prefix(self):
        created_subnetpool = self._create_subnetpool()
        req_body = {'prefixes': self.prefixes_to_add}

        # Add a prefix to the subnet pool
        resp = self.client.add_subnetpool_prefix(created_subnetpool['id'],
                                                 **req_body)
        self._compare_prefix_lists(self.prefixes + self.prefixes_to_add,
                                   resp['prefixes'])

        # Remove the prefix from the subnet pool
        resp = self.client.remove_subnetpool_prefix(created_subnetpool['id'],
                                                    **req_body)
        self._compare_prefix_lists(self.prefixes, resp['prefixes'])

    @decorators.idempotent_id('a36c18fc-10b5-4ebc-ab79-914f826c5bf5')
    @utils.requires_ext(extension='subnetpool-prefix-ops', service='network')
    def test_add_overlapping_prefix(self):
        created_subnetpool = self._create_subnetpool()
        req_body = {'prefixes': self.overlapping_prefixes}

        # Add an overlapping prefix to the subnet pool
        resp = self.client.add_subnetpool_prefix(created_subnetpool['id'],
                                                 **req_body)
        self._compare_prefix_lists(self.prefixes + self.overlapping_prefixes,
                                   resp['prefixes'])


class SubnetPoolPrefixOpsIpv4Test(test_subnetpools.SubnetPoolsTestBase,
                                  SubnetPoolPrefixOpsTestMixin):

    prefixes = ['192.168.1.0/24', '10.10.10.0/24']
    prefixes_to_add = ['192.168.2.0/24']
    overlapping_prefixes = ['10.10.0.0/16']
    min_prefixlen = 16
    ip_version = 4

    @classmethod
    def resource_setup(cls):
        super(SubnetPoolPrefixOpsIpv4Test, cls).resource_setup()
        cls._subnetpool_data = {'prefixes': cls.prefixes,
                                'min_prefixlen': cls.min_prefixlen}


class SubnetPoolPrefixOpsIpv6Test(test_subnetpools.SubnetPoolsTestBase,
                                  SubnetPoolPrefixOpsTestMixin):

    prefixes = ['2001:db8:1234::/48', '2001:db8:1235::/48']
    prefixes_to_add = ['2001:db8:4321::/48']
    overlapping_prefixes = ['2001:db8:1234:1111::/64']
    min_prefixlen = 48
    ip_version = 6

    @classmethod
    def resource_setup(cls):
        super(SubnetPoolPrefixOpsIpv6Test, cls).resource_setup()
        cls._subnetpool_data = {'prefixes': cls.prefixes,
                                'min_prefixlen': cls.min_prefixlen}



# Copyright 2015 Hewlett-Packard Development Company, L.P.dsvsv
# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc
import testtools

from neutron_tempest_plugin.api import base


class SharedNetworksTest(base.BaseAdminNetworkTest):

    @classmethod
    def resource_setup(cls):
        super(SharedNetworksTest, cls).resource_setup()
        cls.shared_network = cls.create_network(shared=True)

    @decorators.idempotent_id('6661d219-b96d-4597-ad10-55766123421a')
    def test_filtering_shared_networks(self):
        # this test is necessary because the 'shared' column does not actually
        # exist on networks so the filter function has to translate it into
        # queries against the RBAC table
        self.create_network()
        self._check_shared_correct(
            self.client.list_networks(shared=True)['networks'], True)
        self._check_shared_correct(
            self.admin_client.list_networks(shared=True)['networks'], True)
        self._check_shared_correct(
            self.client.list_networks(shared=False)['networks'], False)
        self._check_shared_correct(
            self.admin_client.list_networks(shared=False)['networks'], False)

    def _check_shared_correct(self, items, shared):
        self.assertNotEmpty(items)
        self.assertTrue(all(n['shared'] == shared for n in items))

    def _list_subnets_ids(self, client, shared):
        body = client.list_subnets(shared=shared)
        return [subnet['id'] for subnet in body['subnets']]

    @decorators.idempotent_id('6661d219-b96d-4597-ad10-51672353421a')
    def test_filtering_shared_subnets(self):
        # shared subnets need to be tested because their shared status isn't
        # visible as a regular API attribute and it's solely dependent on the
        # parent network
        reg = self.create_network()
        priv = self.create_subnet(reg, client=self.client)
        shared = self.create_subnet(self.shared_network,
                                    client=self.admin_client)
        self.assertIn(shared['id'],
                      self._list_subnets_ids(self.client, shared=True))
        self.assertIn(shared['id'],
                      self._list_subnets_ids(self.admin_client, shared=True))
        self.assertNotIn(priv['id'],
                         self._list_subnets_ids(self.client, shared=True))
        self.assertNotIn(
            priv['id'],
            self._list_subnets_ids(self.admin_client, shared=True))
        self.assertIn(priv['id'],
                      self._list_subnets_ids(self.client, shared=False))
        self.assertIn(priv['id'],
                      self._list_subnets_ids(self.admin_client, shared=False))
        self.assertNotIn(shared['id'],
                         self._list_subnets_ids(self.client, shared=False))
        self.assertNotIn(
            shared['id'],
            self._list_subnets_ids(self.admin_client, shared=False))

    @decorators.idempotent_id('6661d219-b96d-4597-ad10-55766ce4abf7')
    def test_create_update_shared_network(self):
        shared_network = self.create_network(shared=True)
        net_id = shared_network['id']
        self.assertEqual('ACTIVE', shared_network['status'])
        self.assertIsNotNone(shared_network['id'])
        self.assertTrue(self.shared_network['shared'])
        new_name = "New_shared_network"
        body = self.admin_client.update_network(net_id, name=new_name,
                                                admin_state_up=False,
                                                shared=False)
        updated_net = body['network']
        self.assertEqual(new_name, updated_net['name'])
        self.assertFalse(updated_net['shared'])
        self.assertFalse(updated_net['admin_state_up'])

    @decorators.idempotent_id('9c31fabb-0181-464f-9ace-95144fe9ca77')
    def test_create_port_shared_network_as_non_admin_tenant(self):
        # create a port as non admin
        port = self.create_port(self.shared_network)
        self.addCleanup(self.admin_client.delete_port, port['id'])
        # verify the tenant id of admin network and non admin port
        self.assertNotEqual(self.shared_network['project_id'],
                            port['project_id'])

    @decorators.idempotent_id('3e39c4a6-9caf-4710-88f1-d20073c6dd76')
    def test_create_bulk_shared_network(self):
        # Creates 2 networks in one request
        net_nm = [data_utils.rand_name('network'),
                  data_utils.rand_name('network')]
        body = self.admin_client.create_bulk_network(net_nm, shared=True)
        created_networks = body['networks']
        for net in created_networks:
            self.addCleanup(self.admin_client.delete_network, net['id'])
            self.assertIsNotNone(net['id'])
            self.assertTrue(net['shared'])

    def _list_shared_networks(self, user):
        body = user.list_networks(shared=True)
        networks_list = [net['id'] for net in body['networks']]
        self.assertIn(self.shared_network['id'], networks_list)
        self.assertTrue(self.shared_network['shared'])

    @decorators.idempotent_id('a064a9fd-e02f-474a-8159-f828cd636a28')
    def test_list_shared_networks(self):
        # List the shared networks and confirm that
        # shared network extension attribute is returned for those networks
        # that are created as shared
        self._list_shared_networks(self.admin_client)
        self._list_shared_networks(self.client)

    def _show_shared_network(self, user):
        body = user.show_network(self.shared_network['id'])
        show_shared_net = body['network']
        self.assertEqual(self.shared_network['name'], show_shared_net['name'])
        self.assertEqual(self.shared_network['id'], show_shared_net['id'])
        self.assertTrue(show_shared_net['shared'])

    @decorators.idempotent_id('e03c92a2-638d-4bfa-b50a-b1f66f087e58')
    def test_show_shared_networks_attribute(self):
        # Show a shared network and confirm that
        # shared network extension attribute is returned.
        self._show_shared_network(self.admin_client)
        self._show_shared_network(self.client)


class AllowedAddressPairSharedNetworkTest(base.BaseAdminNetworkTest):
    allowed_address_pairs = [{'ip_address': '1.1.1.1'}]
    required_extensions = ['allowed-address-pairs']

    @classmethod
    def resource_setup(cls):
        super(AllowedAddressPairSharedNetworkTest, cls).resource_setup()
        cls.network = cls.create_network(shared=True)
        cls.create_subnet(cls.network, client=cls.admin_client)

    @decorators.idempotent_id('86c3529b-1231-40de-803c-ffffffff1fff')
    def test_create_with_address_pair_blocked_on_other_network(self):
        with testtools.ExpectedException(lib_exc.Forbidden):
            self.create_port(self.network,
                             allowed_address_pairs=self.allowed_address_pairs)

    @decorators.idempotent_id('86c3529b-1231-40de-803c-ffffffff2fff')
    def test_update_with_address_pair_blocked_on_other_network(self):
        port = self.create_port(self.network)
        with testtools.ExpectedException(lib_exc.Forbidden):
            self.update_port(
                port, allowed_address_pairs=self.allowed_address_pairs)


class RBACSharedNetworksTest(base.BaseAdminNetworkTest):

    force_tenant_isolation = True
    credentials = ['primary', 'alt', 'admin']
    required_extensions = ['rbac-policies']

    @classmethod
    def resource_setup(cls):
        super(RBACSharedNetworksTest, cls).resource_setup()
        cls.client2 = cls.os_alt.network_client

    def _make_admin_net_and_subnet_shared_to_project_id(self, project_id):
        net = self.admin_client.create_network(
            name=data_utils.rand_name('test-network'))['network']
        self.addCleanup(self.admin_client.delete_network, net['id'])
        subnet = self.create_subnet(net, client=self.admin_client)
        # network is shared to first unprivileged client by default
        pol = self.admin_client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared', target_tenant=project_id
        )['rbac_policy']
        return {'network': net, 'subnet': subnet, 'policy': pol}

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('86c3529b-1231-40de-803c-bfffffff1eee')
    def test_create_rbac_policy_with_target_tenant_none(self):
        with testtools.ExpectedException(lib_exc.BadRequest):
            self._make_admin_net_and_subnet_shared_to_project_id(
                project_id=None)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('86c3529b-1231-40de-803c-bfffffff1fff')
    def test_create_rbac_policy_with_target_tenant_too_long_id(self):
        with testtools.ExpectedException(lib_exc.BadRequest):
            target_project = '1234' * 100
            self._make_admin_net_and_subnet_shared_to_project_id(
                project_id=target_project)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff1fff')
    def test_network_only_visible_to_policy_target(self):
        net = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)['network']
        self.client.show_network(net['id'])
        with testtools.ExpectedException(lib_exc.NotFound):
            # client2 has not been granted access
            self.client2.show_network(net['id'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff2fff')
    def test_subnet_on_network_only_visible_to_policy_target(self):
        sub = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)['subnet']
        self.client.show_subnet(sub['id'])
        with testtools.ExpectedException(lib_exc.NotFound):
            # client2 has not been granted access
            self.client2.show_subnet(sub['id'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff2eee')
    def test_policy_target_update(self):
        res = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        # change to client2
        update_res = self.admin_client.update_rbac_policy(
                res['policy']['id'], target_tenant=self.client2.project_id)
        self.assertEqual(self.client2.project_id,
                         update_res['rbac_policy']['target_tenant'])
        # make sure everything else stayed the same
        res['policy'].pop('target_tenant')
        update_res['rbac_policy'].pop('target_tenant')
        self.assertEqual(res['policy'], update_res['rbac_policy'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-affefefef321')
    def test_duplicate_policy_error(self):
        res = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        with testtools.ExpectedException(lib_exc.Conflict):
            self.admin_client.create_rbac_policy(
                object_type='network', object_id=res['network']['id'],
                action='access_as_shared',
                target_tenant=self.client.project_id)

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff3fff')
    def test_port_presence_prevents_network_rbac_policy_deletion(self):
        res = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        port = self.create_port(res['network'])
        # a port on the network should prevent the deletion of a policy
        # required for it to exist
        with testtools.ExpectedException(lib_exc.Conflict):
            self.admin_client.delete_rbac_policy(res['policy']['id'])

        self.client.delete_port(port['id'])
        # anchor is gone, delete should pass
        self.admin_client.delete_rbac_policy(res['policy']['id'])

    @decorators.idempotent_id('34d627da-a732-68c0-2e1a-bc4a19246698')
    def test_delete_self_share_rule(self):
        net = self.create_network()
        self_share = self.client.create_rbac_policy(
                         object_type='network', object_id=net['id'],
                         action='access_as_shared',
                         target_tenant=net['project_id'])['rbac_policy']
        port = self.create_port(net)
        self.client.delete_rbac_policy(self_share['id'])
        self.client.delete_port(port['id'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-beefbeefbeef')
    def test_tenant_can_delete_port_on_own_network(self):
        net = self.create_network()  # owned by self.client
        self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared', target_tenant=self.client2.project_id)
        port = self.client2.create_port(network_id=net['id'])['port']
        self.client.delete_port(port['id'])

    @decorators.idempotent_id('f7539232-389a-4e9c-9e37-e42a129eb541')
    def test_tenant_cant_delete_other_tenants_ports(self):
        net = self.create_network()
        port = self.create_port(net)
        with testtools.ExpectedException(lib_exc.NotFound):
            self.client2.delete_port(port['id'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff4fff')
    def test_regular_client_shares_to_another_regular_client(self):
        net = self.create_network()  # owned by self.client
        with testtools.ExpectedException(lib_exc.NotFound):
            self.client2.show_network(net['id'])
        pol = self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared', target_tenant=self.client2.project_id)
        self.client2.show_network(net['id'])

        self.assertIn(pol['rbac_policy'],
                      self.client.list_rbac_policies()['rbac_policies'])
        # ensure that 'client2' can't see the policy sharing the network to it
        # because the policy belongs to 'client'
        self.assertNotIn(pol['rbac_policy']['id'],
            [p['id']
             for p in self.client2.list_rbac_policies()['rbac_policies']])

    @decorators.idempotent_id('bf5052b8-b11e-407c-8e43-113447404d3e')
    def test_filter_fields(self):
        net = self.create_network()
        self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared', target_tenant=self.client2.project_id)
        field_args = (('id',), ('id', 'action'), ('object_type', 'object_id'),
                      ('project_id', 'target_tenant'))
        for fields in field_args:
            res = self.client.list_rbac_policies(fields=fields)
            self.assertEqual(set(fields), set(res['rbac_policies'][0].keys()))

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff5fff')
    def test_policy_show(self):
        res = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        p1 = res['policy']
        p2 = self.admin_client.create_rbac_policy(
            object_type='network', object_id=res['network']['id'],
            action='access_as_shared',
            target_tenant='*')['rbac_policy']

        self.assertEqual(
            p1, self.admin_client.show_rbac_policy(p1['id'])['rbac_policy'])
        self.assertEqual(
            p2, self.admin_client.show_rbac_policy(p2['id'])['rbac_policy'])

    @decorators.idempotent_id('e7bcb1ea-4877-4266-87bb-76f68b421f31')
    def test_filter_policies(self):
        net = self.create_network()
        pol1 = self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared',
            target_tenant=self.client2.project_id)['rbac_policy']
        pol2 = self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared',
            target_tenant=self.client.project_id)['rbac_policy']
        res1 = self.client.list_rbac_policies(id=pol1['id'])['rbac_policies']
        res2 = self.client.list_rbac_policies(id=pol2['id'])['rbac_policies']
        self.assertEqual(1, len(res1))
        self.assertEqual(1, len(res2))
        self.assertEqual(pol1['id'], res1[0]['id'])
        self.assertEqual(pol2['id'], res2[0]['id'])

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff6fff')
    def test_regular_client_blocked_from_sharing_anothers_network(self):
        net = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)['network']
        with testtools.ExpectedException(lib_exc.BadRequest):
            self.client.create_rbac_policy(
                object_type='network', object_id=net['id'],
                action='access_as_shared',
                target_tenant=self.client.project_id)

    @decorators.idempotent_id('c5f8f785-ce8d-4430-af7e-a236205862fb')
    @utils.requires_ext(extension="quotas", service="network")
    def test_rbac_policy_quota(self):
        quota = self.client.show_quotas(self.client.project_id)['quota']
        max_policies = quota['rbac_policy']
        self.assertGreater(max_policies, 0)
        net = self.client.create_network(
            name=data_utils.rand_name('test-network'))['network']
        self.addCleanup(self.client.delete_network, net['id'])
        with testtools.ExpectedException(lib_exc.Conflict):
            for i in range(0, max_policies + 1):
                self.admin_client.create_rbac_policy(
                    object_type='network', object_id=net['id'],
                    action='access_as_shared',
                    target_tenant=uuidutils.generate_uuid().replace('-', ''))

    @decorators.idempotent_id('86c3529b-1231-40de-803c-afffffff7fff')
    def test_regular_client_blocked_from_sharing_with_wildcard(self):
        net = self.create_network()
        with testtools.ExpectedException(lib_exc.Forbidden):
            self.client.create_rbac_policy(
                object_type='network', object_id=net['id'],
                action='access_as_shared', target_tenant='*')
        # ensure it works on update as well
        pol = self.client.create_rbac_policy(
            object_type='network', object_id=net['id'],
            action='access_as_shared', target_tenant=self.client2.project_id)
        with testtools.ExpectedException(lib_exc.Forbidden):
            self.client.update_rbac_policy(pol['rbac_policy']['id'],
                                           target_tenant='*')

    @decorators.idempotent_id('34d627da-869f-68c0-2e1a-bc4a19246698')
    def test_update_self_share_rule(self):
        net = self.create_network()
        self_share = self.client.create_rbac_policy(
                         object_type='network', object_id=net['id'],
                         action='access_as_shared',
                         target_tenant=net['project_id'])['rbac_policy']
        port = self.create_port(net)
        self.client.update_rbac_policy(self_share['id'],
                                       target_tenant=self.client2.project_id)
        self.client.delete_port(port['id'])

    @utils.requires_ext(extension="standard-attr-revisions", service="network")
    @decorators.idempotent_id('86c3529b-1231-40de-1234-89664291a4cb')
    def test_rbac_bumps_network_revision(self):
        resp = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        net_id = resp['network']['id']
        rev = self.client.show_network(net_id)['network']['revision_number']
        self.admin_client.create_rbac_policy(
            object_type='network', object_id=net_id,
            action='access_as_shared', target_tenant='*')
        self.assertGreater(
            self.client.show_network(net_id)['network']['revision_number'],
            rev
        )

    @decorators.idempotent_id('86c3529b-1231-40de-803c-aeeeeeee7fff')
    def test_filtering_works_with_rbac_records_present(self):
        resp = self._make_admin_net_and_subnet_shared_to_project_id(
            self.client.project_id)
        net = resp['network']['id']
        sub = resp['subnet']['id']
        self.admin_client.create_rbac_policy(
            object_type='network', object_id=net,
            action='access_as_shared', target_tenant='*')
        self._assert_shared_object_id_listing_presence('subnets', False, sub)
        self._assert_shared_object_id_listing_presence('subnets', True, sub)
        self._assert_shared_object_id_listing_presence('networks', False, net)
        self._assert_shared_object_id_listing_presence('networks', True, net)

    def _assert_shared_object_id_listing_presence(self, resource, shared, oid):
        lister = getattr(self.admin_client, 'list_%s' % resource)
        objects = [o['id'] for o in lister(shared=shared)[resource]]
        if shared:
            self.assertIn(oid, objects)
        else:
            self.assertNotIn(oid, objects)



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import subprocess

from debtcollector import removals
import netaddr
from neutron_lib.api import validators
from neutron_lib import constants as neutron_lib_constants
from oslo_log import log
from packaging import version as packaging_version
from paramiko import ssh_exception as ssh_exc
from tempest.common.utils import net_utils
from tempest.common import waiters
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import base as base_api
from neutron_tempest_plugin.common import ip as ip_utils
from neutron_tempest_plugin.common import shell
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin import exceptions
from neutron_tempest_plugin.scenario import constants

CONF = config.CONF
LOG = log.getLogger(__name__)
SSH_EXC_TUPLE = (lib_exc.SSHTimeout,
                 ssh_exc.AuthenticationException,
                 ssh_exc.NoValidConnectionsError,
                 ConnectionResetError)


def get_ncat_version(ssh_client=None):
    cmd = "ncat --version 2>&1"
    try:
        version_result = shell.execute(cmd, ssh_client=ssh_client).stdout
    except exceptions.ShellCommandFailed:
        m = None
    else:
        m = re.match(r"Ncat: Version ([\d.]+) *.", version_result)
    # NOTE(slaweq): by default lets assume we have ncat 7.60 which is in Ubuntu
    # 18.04 which is used on u/s gates
    return packaging_version.Version(m.group(1) if m else '7.60')


def get_ncat_server_cmd(port, protocol, msg=None):
    udp = ''
    if protocol.lower() == neutron_lib_constants.PROTO_NAME_UDP:
        udp = '-u'
    cmd = "nc %(udp)s -p %(port)s -lk " % {
        'udp': udp, 'port': port}
    if msg:
        if CONF.neutron_plugin_options.default_image_is_advanced:
            cmd += "-c 'echo %s' " % msg
        else:
            cmd += "-e echo %s " % msg
    cmd += "< /dev/zero &{0}sleep 0.1{0}".format('\n')
    return cmd


def get_ncat_client_cmd(ip_address, port, protocol, ssh_client=None):
    cmd = 'echo "knock knock" | nc '
    ncat_version = get_ncat_version(ssh_client=ssh_client)
    if ncat_version > packaging_version.Version('7.60'):
        cmd += '-d 1 '
    if protocol.lower() == neutron_lib_constants.PROTO_NAME_UDP:
        cmd += '-u '
        if ncat_version > packaging_version.Version('7.60'):
            cmd += '-z '
    cmd += '-w 1 %(host)s %(port)s' % {'host': ip_address, 'port': port}
    return cmd


class BaseTempestTestCase(base_api.BaseNetworkTest):

    def create_server(self, flavor_ref, image_ref, key_name, networks,
                      **kwargs):
        """Create a server using tempest lib

        All the parameters are the ones used in Compute API
        * - Kwargs that require admin privileges

        Args:
           flavor_ref(str): The flavor of the server to be provisioned.
           image_ref(str):  The image of the server to be provisioned.
           key_name(str): SSH key to to be used to connect to the
                            provisioned server.
           networks(list): List of dictionaries where each represent
               an interface to be attached to the server. For network
               it should be {'uuid': network_uuid} and for port it should
               be {'port': port_uuid}
        kwargs:
           name(str): Name of the server to be provisioned.
           security_groups(list): List of dictionaries where
                the keys is 'name' and the value is the name of
                the security group. If it's not passed the default
                security group will be used.
           availability_zone(str)*: The availability zone that
                the instance will be in.
                You can request a specific az without actually creating one,
                Just pass 'X:Y' where X is the default availability
                zone, and Y is the compute host name.
        """

        kwargs.setdefault('name', data_utils.rand_name('server-test'))

        # We cannot use setdefault() here because caller could have passed
        # security_groups=None and we don't want to pass None to
        # client.create_server()
        if not kwargs.get('security_groups'):
            kwargs['security_groups'] = [{'name': 'default'}]

        client = kwargs.pop('client', None)
        if client is None:
            client = self.os_primary.servers_client
            if kwargs.get('availability_zone'):
                client = self.os_admin.servers_client

        server = client.create_server(
            flavorRef=flavor_ref,
            imageRef=image_ref,
            key_name=key_name,
            networks=networks,
            **kwargs)

        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        waiters.wait_for_server_termination,
                        client,
                        server['server']['id'])
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        client.delete_server,
                        server['server']['id'])

        self.wait_for_server_active(server['server'], client=client)
        self.wait_for_guest_os_ready(server['server'], client=client)

        return server

    @classmethod
    def create_secgroup_rules(cls, rule_list, secgroup_id=None,
                              client=None):
        client = client or cls.os_primary.network_client
        if not secgroup_id:
            sgs = client.list_security_groups()['security_groups']
            for sg in sgs:
                if sg['name'] == constants.DEFAULT_SECURITY_GROUP:
                    secgroup_id = sg['id']
                    break
        resp = []
        for rule in rule_list:
            direction = rule.pop('direction')
            resp.append(client.create_security_group_rule(
                        direction=direction,
                        security_group_id=secgroup_id,
                        **rule)['security_group_rule'])
        return resp

    @classmethod
    def create_loginable_secgroup_rule(cls, secgroup_id=None,
                                       client=None):
        """This rule is intended to permit inbound ssh

        Allowing ssh traffic traffic from all sources, so no group_id is
        provided.
        Setting a group_id would only permit traffic from ports
        belonging to the same security group.
        """
        return cls.create_security_group_rule(
            security_group_id=secgroup_id,
            client=client,
            protocol=neutron_lib_constants.PROTO_NAME_TCP,
            direction=neutron_lib_constants.INGRESS_DIRECTION,
            port_range_min=22,
            port_range_max=22)

    @classmethod
    def create_ingress_metadata_secgroup_rule(cls, secgroup_id=None):
        """This rule is intended to permit inbound metadata traffic

        Allowing ingress traffic from metadata server, required only for
        stateless security groups.
        """
        # NOTE(slaweq): in case of stateless security groups, there is no
        # "related" or "established" traffic matching at all so even if
        # egress traffic to 169.254.169.254 is allowed by default SG, we
        # need to explicitly allow ingress traffic from the metadata server
        # to be able to receive responses in the guest vm
        cls.create_security_group_rule(
            security_group_id=secgroup_id,
            direction=neutron_lib_constants.INGRESS_DIRECTION,
            protocol=neutron_lib_constants.PROTO_NAME_TCP,
            remote_ip_prefix='169.254.169.254/32',
            description='metadata out'
        )

    @classmethod
    def create_pingable_secgroup_rule(cls, secgroup_id=None,
                                      client=None):
        """This rule is intended to permit inbound ping

        """
        return cls.create_security_group_rule(
            security_group_id=secgroup_id, client=client,
            protocol=neutron_lib_constants.PROTO_NAME_ICMP,
            direction=neutron_lib_constants.INGRESS_DIRECTION)

    @classmethod
    def create_router_by_client(cls, is_admin=False, **kwargs):
        kwargs.update({'router_name': data_utils.rand_name('router'),
                       'admin_state_up': True,
                       'external_network_id': CONF.network.public_network_id})
        if not is_admin:
            router = cls.create_router(**kwargs)
        else:
            router = cls.create_admin_router(**kwargs)
        LOG.debug("Created router %s", router['name'])
        cls._wait_for_router_ha_active(router['id'])
        return router

    @classmethod
    def _wait_for_router_ha_active(cls, router_id):
        router = cls.os_admin.network_client.show_router(router_id)['router']
        if not router.get('ha') or cls.is_driver_ovn:
            return

        def _router_active_on_l3_agent():
            agents = cls.os_admin.network_client.list_l3_agents_hosting_router(
                router_id)['agents']
            return "active" in [agent['ha_state'] for agent in agents]

        error_msg = (
            "Router %s is not active on any of the L3 agents" % router_id)
        # NOTE(slaweq): timeout here should be lower for sure, but due to
        # the bug https://launchpad.net/bugs/1923633 let's wait even 10
        # minutes until router will be active on some of the L3 agents
        utils.wait_until_true(_router_active_on_l3_agent,
                              timeout=600, sleep=5,
                              exception=lib_exc.TimeoutException(error_msg))

    @classmethod
    def skip_if_no_extension_enabled_in_l3_agents(cls, extension):
        l3_agents = cls.os_admin.network_client.list_agents(
                binary='neutron-l3-agent')['agents']
        if not l3_agents:
            # the tests should not be skipped when neutron-l3-agent does not
            # exist (this validation doesn't apply to the setups like
            # e.g. ML2/OVN)
            return
        for agent in l3_agents:
            if extension in agent['configurations'].get('extensions', []):
                return
        raise cls.skipTest("No L3 agent with '%s' extension enabled found." %
                           extension)

    @removals.remove(version='Stein',
                     message="Please use create_floatingip method instead of "
                             "create_and_associate_floatingip.")
    def create_and_associate_floatingip(self, port_id, client=None):
        client = client or self.os_primary.network_client
        return self.create_floatingip(port_id=port_id, client=client)

    def create_interface(cls, server_id, port_id, client=None):
        client = client or cls.os_primary.interfaces_client
        body = client.create_interface(server_id, port_id=port_id)
        return body['interfaceAttachment']

    def delete_interface(cls, server_id, port_id, client=None):
        client = client or cls.os_primary.interfaces_client
        client.delete_interface(server_id, port_id=port_id)

    def setup_network_and_server(self, router=None, server_name=None,
                                 network=None, use_stateless_sg=False,
                                 **kwargs):
        """Create network resources and a server.

        Creating a network, subnet, router, keypair, security group
        and a server.
        """
        self.network = network or self.create_network()
        LOG.debug("Created network %s", self.network['name'])
        self.subnet = self.create_subnet(self.network)
        LOG.debug("Created subnet %s", self.subnet['id'])

        sg_args = {
            'name': data_utils.rand_name('secgroup')
        }
        if use_stateless_sg:
            sg_args['stateful'] = False
        secgroup = self.os_primary.network_client.create_security_group(
            **sg_args)
        LOG.debug("Created security group %s",
                  secgroup['security_group']['name'])
        self.security_groups.append(secgroup['security_group'])
        if not router:
            router = self.create_router_by_client(**kwargs)
        self.create_router_interface(router['id'], self.subnet['id'])
        self.keypair = self.create_keypair()
        self.create_loginable_secgroup_rule(
            secgroup_id=secgroup['security_group']['id'])
        if use_stateless_sg:
            self.create_ingress_metadata_secgroup_rule(
                secgroup_id=secgroup['security_group']['id'])

        server_kwargs = {
            'flavor_ref': CONF.compute.flavor_ref,
            'image_ref': CONF.compute.image_ref,
            'key_name': self.keypair['name'],
            'networks': [{'uuid': self.network['id']}],
            'security_groups': [{'name': secgroup['security_group']['name']}],
        }
        if server_name is not None:
            server_kwargs['name'] = server_name

        self.server = self.create_server(**server_kwargs)
        self.port = self.client.list_ports(network_id=self.network['id'],
                                           device_id=self.server[
                                               'server']['id'])['ports'][0]
        self.fip = self.create_floatingip(port=self.port)

    def check_connectivity(self, host, ssh_user=None, ssh_key=None,
                           servers=None, ssh_timeout=None, ssh_client=None):
        # Either ssh_client or ssh_user+ssh_key is mandatory.
        if ssh_client is None:
            ssh_client = ssh.Client(host, ssh_user,
                                    pkey=ssh_key, timeout=ssh_timeout)
        try:
            ssh_client.test_connection_auth()
        except SSH_EXC_TUPLE as ssh_e:
            LOG.debug(ssh_e)
            self._log_console_output(servers)
            self._log_local_network_status()
            raise

    def _log_console_output(self, servers=None):
        if not CONF.compute_feature_enabled.console_output:
            LOG.debug('Console output not supported, cannot log')
            return
        if not servers:
            servers = self.os_primary.servers_client.list_servers()
            servers = servers['servers']
        for server in servers:
            # NOTE(slaweq): sometimes servers are passed in dictionary with
            # "server" key as first level key and in other cases it may be that
            # it is just the "inner" dict without "server" key. Lets try to
            # handle both cases
            server = server.get("server") or server
            try:
                console_output = (
                    self.os_primary.servers_client.get_console_output(
                        server['id'])['output'])
                LOG.debug('Console output for %s\nbody=\n%s',
                          server['id'], console_output)
            except lib_exc.NotFound:
                LOG.debug("Server %s disappeared(deleted) while looking "
                          "for the console log", server['id'])

    def _log_local_network_status(self):
        self._log_ns_network_status()
        for ns_name in ip_utils.IPCommand().list_namespaces():
            self._log_ns_network_status(ns_name=ns_name)

    def _log_ns_network_status(self, ns_name=None):
        try:
            local_ips = ip_utils.IPCommand(namespace=ns_name).list_addresses()
            local_routes = ip_utils.IPCommand(namespace=ns_name).list_routes()
            arp_table = ip_utils.arp_table(namespace=ns_name)
            iptables = ip_utils.list_iptables(namespace=ns_name)
            lsockets = ip_utils.list_listening_sockets(namespace=ns_name)
        except exceptions.ShellCommandFailed:
            LOG.debug('Namespace %s has been deleted synchronously during the '
                      'host network collection process', ns_name)
            return

        LOG.debug('Namespace %s; IP Addresses:\n%s',
                  ns_name, '\n'.join(str(r) for r in local_ips))
        LOG.debug('Namespace %s; Local routes:\n%s',
                  ns_name, '\n'.join(str(r) for r in local_routes))
        LOG.debug('Namespace %s; Local ARP table:\n%s',
                  ns_name, '\n'.join(str(r) for r in arp_table))
        LOG.debug('Namespace %s; Local iptables:\n%s', ns_name, iptables)
        LOG.debug('Namespace %s; Listening sockets:\n%s', ns_name, lsockets)

    def _check_remote_connectivity(self, source, dest, count,
                                   should_succeed=True,
                                   nic=None, mtu=None, fragmentation=True,
                                   timeout=None, pattern=None,
                                   forbid_packet_loss=False,
                                   check_response_ip=True):
        """check ping server via source ssh connection

        :param source: RemoteClient: an ssh connection from which to ping
        :param dest: and IP to ping against
        :param count: Number of ping packet(s) to send
        :param should_succeed: boolean should ping succeed or not
        :param nic: specific network interface to ping from
        :param mtu: mtu size for the packet to be sent
        :param fragmentation: Flag for packet fragmentation
        :param timeout: Timeout for all ping packet(s) to succeed
        :param pattern: hex digits included in ICMP messages
        :param forbid_packet_loss: forbid or allow some lost packets
        :param check_response_ip: check response ip
        :returns: boolean -- should_succeed == ping
        :returns: ping is false if ping failed
        """
        def ping_host(source, host, count,
                      size=CONF.validation.ping_size, nic=None, mtu=None,
                      fragmentation=True, pattern=None):
            IP_VERSION_4 = neutron_lib_constants.IP_VERSION_4
            IP_VERSION_6 = neutron_lib_constants.IP_VERSION_6

            # Use 'ping6' for IPv6 addresses, 'ping' for IPv4 and hostnames
            ip_version = (
                IP_VERSION_6 if netaddr.valid_ipv6(host) else IP_VERSION_4)
            cmd = (
                'ping6' if ip_version == IP_VERSION_6 else 'ping')
            if nic:
                cmd = 'sudo {cmd} -I {nic}'.format(cmd=cmd, nic=nic)
            if mtu:
                if not fragmentation:
                    cmd += ' -M do'
                size = str(net_utils.get_ping_payload_size(
                    mtu=mtu, ip_version=ip_version))
            if pattern:
                cmd += ' -p {pattern}'.format(pattern=pattern)
            cmd += ' -c{0} -W{0} -s{1} {2}'.format(count, size, host)
            return source.exec_command(cmd)

        def ping_remote():
            try:
                result = ping_host(source, dest, count, nic=nic, mtu=mtu,
                                   fragmentation=fragmentation,
                                   pattern=pattern)

            except lib_exc.SSHExecCommandFailed:
                LOG.warning('Failed to ping IP: %s via a ssh connection '
                            'from: %s.', dest, source.host)
                return not should_succeed
            LOG.debug('ping result: %s', result)

            if forbid_packet_loss and ' 0% packet loss' not in result:
                LOG.debug('Packet loss detected')
                return not should_succeed

            if (check_response_ip and
                    validators.validate_ip_address(dest) is None):
                # Assert that the return traffic was from the correct
                # source address.
                from_source = 'from %s' % dest
                self.assertIn(from_source, result)
            return should_succeed

        return test_utils.call_until_true(
            ping_remote, timeout or CONF.validation.ping_timeout, 1)

    def check_remote_connectivity(self, source, dest, should_succeed=True,
                                  nic=None, mtu=None, fragmentation=True,
                                  servers=None, timeout=None,
                                  ping_count=CONF.validation.ping_count,
                                  pattern=None, forbid_packet_loss=False,
                                  check_response_ip=True):
        try:
            self.assertTrue(self._check_remote_connectivity(
                source, dest, ping_count, should_succeed, nic, mtu,
                fragmentation,
                timeout=timeout, pattern=pattern,
                forbid_packet_loss=forbid_packet_loss,
                check_response_ip=check_response_ip))
        except SSH_EXC_TUPLE as ssh_e:
            LOG.debug(ssh_e)
            self._log_console_output(servers)
            self._log_local_network_status()
            raise
        except AssertionError:
            self._log_console_output(servers)
            self._log_local_network_status()
            raise

    def ping_ip_address(self, ip_address, should_succeed=True,
                        ping_timeout=None, mtu=None):
        # the code is taken from tempest/scenario/manager.py in tempest git
        timeout = ping_timeout or CONF.validation.ping_timeout
        cmd = ['ping', '-c1', '-w1']

        if mtu:
            cmd += [
                # don't fragment
                '-M', 'do',
                # ping receives just the size of ICMP payload
                '-s', str(net_utils.get_ping_payload_size(mtu, 4))
            ]
        cmd.append(ip_address)

        def ping():
            proc = subprocess.Popen(cmd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            proc.communicate()

            return (proc.returncode == 0) == should_succeed

        caller = test_utils.find_test_caller()
        LOG.debug('%(caller)s begins to ping %(ip)s in %(timeout)s sec and the'
                  ' expected result is %(should_succeed)s', {
                      'caller': caller, 'ip': ip_address, 'timeout': timeout,
                      'should_succeed':
                      'reachable' if should_succeed else 'unreachable'
                  })
        result = test_utils.call_until_true(ping, timeout, 1)

        # To make sure ping_ip_address called by test works
        # as expected.
        self.assertTrue(result)

        LOG.debug('%(caller)s finishes ping %(ip)s in %(timeout)s sec and the '
                  'ping result is %(result)s', {
                      'caller': caller, 'ip': ip_address, 'timeout': timeout,
                      'result': 'expected' if result else 'unexpected'
                  })
        return result

    def wait_for_server_status(self, server, status, client=None, **kwargs):
        """Waits for a server to reach a given status.

        :param server:  mapping having schema {'id': <server_id>}
        :param status: string status to wait for (es: 'ACTIVE')
        :param clien:  servers client (self.os_primary.servers_client as
                       default value)
        """

        client = client or self.os_primary.servers_client
        waiters.wait_for_server_status(client, server['id'], status, **kwargs)

    def wait_for_server_active(self, server, client=None):
        """Waits for a server to reach active status.

        :param server:  mapping having schema {'id': <server_id>}
        :param clien:  servers client (self.os_primary.servers_client as
                       default value)
        """
        self.wait_for_server_status(
            server, constants.SERVER_STATUS_ACTIVE, client)

    def wait_for_guest_os_ready(self, server, client=None):
        if not CONF.compute_feature_enabled.console_output:
            LOG.debug('Console output not supported, cannot check if server '
                      '%s is ready.', server['id'])
            return

        client = client or self.os_primary.servers_client

        def system_booted():
            console_output = client.get_console_output(server['id'])['output']
            for line in console_output.split('\n'):
                if 'login:' in line.lower():
                    return True
            return False

        try:
            utils.wait_until_true(system_booted, timeout=90, sleep=5)
        except utils.WaitTimeout:
            LOG.debug("No correct output in console of server %s found. "
                      "Guest operating system status can't be checked.",
                      server['id'])

    def check_servers_hostnames(self, servers, timeout=None, log_errors=True,
                                external_port=None):
        """Compare hostnames of given servers with their names."""
        try:
            for server in servers:
                kwargs = {}
                if timeout:
                    kwargs['timeout'] = timeout
                try:
                    kwargs['port'] = external_port or (
                        server['port_forwarding_tcp']['external_port'])
                except KeyError:
                    pass
                ssh_client = ssh.Client(
                    self.fip['floating_ip_address'],
                    CONF.validation.image_ssh_user,
                    pkey=self.keypair['private_key'],
                    **kwargs)
                self.assertIn(server['name'],
                              ssh_client.get_hostname())
        except SSH_EXC_TUPLE as ssh_e:
            LOG.debug(ssh_e)
            if log_errors:
                self._log_console_output(servers)
                self._log_local_network_status()
            raise
        except AssertionError as assert_e:
            LOG.debug(assert_e)
            if log_errors:
                self._log_console_output(servers)
                self._log_local_network_status()
            raise

    def ensure_nc_listen(self, ssh_client, port, protocol, echo_msg=None,
                         servers=None):
        """Ensure that nc server listening on the given TCP/UDP port is up.

        Listener is created always on remote host.
        """
        def spawn_and_check_process():
            self.nc_listen(ssh_client, port, protocol, echo_msg, servers)
            return utils.process_is_running(ssh_client, "nc")

        utils.wait_until_true(spawn_and_check_process)

    def nc_listen(self, ssh_client, port, protocol, echo_msg=None,
                  servers=None):
        """Create nc server listening on the given TCP/UDP port.

        Listener is created always on remote host.
        """
        try:
            return ssh_client.execute_script(
                get_ncat_server_cmd(port, protocol, echo_msg),
                become_root=True, combine_stderr=True)
        except SSH_EXC_TUPLE as ssh_e:
            LOG.debug(ssh_e)
            self._log_console_output(servers)
            self._log_local_network_status()
            raise

    def nc_client(self, ip_address, port, protocol, ssh_client=None):
        """Check connectivity to TCP/UDP port at host via nc.

        If ssh_client is not given, it is executed locally on host where tests
        are executed. Otherwise ssh_client object is used to execute it.
        """
        cmd = get_ncat_client_cmd(ip_address, port, protocol,
                                  ssh_client=ssh_client)
        result = shell.execute(cmd, ssh_client=ssh_client, check=False)
        return result.stdout

    def _ensure_public_router(self, client=None, tenant_id=None):
        """Retrieve a router for the given tenant id.

        If a public router has been configured, it will be returned.

        If a public router has not been configured, but a public
        network has, a tenant router will be created and returned that
        routes traffic to the public network.
        """
        if not client:
            client = self.client
        if not tenant_id:
            tenant_id = client.project_id
        router_id = CONF.network.public_router_id
        network_id = CONF.network.public_network_id
        if router_id:
            body = client.show_router(router_id)
            return body['router']
        elif network_id:
            router = self.create_router_by_client()
            self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                            client.delete_router, router['id'])
            kwargs = {'external_gateway_info': dict(network_id=network_id)}
            router = client.update_router(router['id'], **kwargs)['router']
            return router
        else:
            raise Exception("Neither of 'public_router_id' or "
                            "'public_network_id' has been defined.")

    def _update_router_admin_state(self, router, admin_state_up):
        kwargs = dict(admin_state_up=admin_state_up)
        router = self.client.update_router(
            router['id'], **kwargs)['router']
        self.assertEqual(admin_state_up, router['admin_state_up'])

    def _check_cmd_installed_on_server(self, ssh_client, server, cmd):
        try:
            ssh_client.execute_script('which %s' % cmd)
        except SSH_EXC_TUPLE as ssh_e:
            LOG.debug(ssh_e)
            self._log_console_output([server])
            self._log_local_network_status()
            raise
        except exceptions.SSHScriptFailed:
            raise self.skipException(
                "%s is not available on server %s" % (cmd, server['id']))
